#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Methods for gluing a user's application into the GAE environment."""

import binascii
import contextvars
import functools
import logging
import logging.handlers
import os
import sys
import traceback

from google.appengine.api import namespace_manager
from google.appengine.api.runtime import runtime
from google.appengine.ext.deferred import deferred
from google.appengine.runtime import background
from google.appengine.runtime import callback
from google.appengine.runtime import context
from google.appengine.runtime import default_api_stub
from google.appengine.runtime import request_environment
import six
from six.moves import urllib



MAX_CONCURRENT_REQUESTS = 501



ENV_VARS_FROM_HTTP_X_APPENGINE_HEADERS = [
    'AUTH_DOMAIN',
    'DATACENTER',
    'DEFAULT_VERSION_HOSTNAME',
    'HTTPS',
    'REMOTE_ADDR',
    'REQUEST_ID_HASH',
    'REQUEST_LOG_ID',
    'USER_EMAIL',
    'USER_ID',
    'USER_IS_ADMIN',
    'USER_NICKNAME',
    'USER_ORGANIZATION',



    'APPSERVER_DATACENTER',
    'APPSERVER_TASK_BNS',
    'GAIA_AUTHUSER',
    'GAIA_ID',
    'GAIA_SESSION',
    'LOAS_PEER_USERNAME',
    'LOAS_SECURITY_LEVEL',
    'TRUSTED_IP_REQUEST',
]

LOCAL_OVERRIDABLE_VARS = [
    'USER_ID',
    'USER_EMAIL',
    'USER_IS_ADMIN',
    'AUTH_DOMAIN',
]


def middleware(f):
  """Function decorator for making WSGI middlewares."""
  return functools.update_wrapper(
      lambda app: lambda wsgi_env, start_resp: f(app, wsgi_env, start_resp),
      f)


def Wrap(app, middlewares):
  """Wrap(app, [a,b,c]) is equivalent to a(b(c(app)))."""
  return functools.reduce(lambda app, mw: mw(app), reversed(middlewares), app)


@middleware
def UseRequestSecurityTicketForApiMiddleware(app, wsgi_env, start_response):
  """WSGI middleware wrapper that sets the thread to use the security ticket.

  This sets up the appengine api so that if a security ticket is passed in with
  the request, it will be used.

  Args:
    app: (callable) a WSGI app per PEP 3333.
    wsgi_env: see PEP 3333
    start_response: see PEP 3333

  Returns:
    A wrapped <app>, which is also a valid WSGI app.
  """
  try:
    default_api_stub.DefaultApiStub.SetUseRequestSecurityTicketForThread(True)
    return app(wsgi_env, start_response)
  finally:

    default_api_stub.DefaultApiStub.SetUseRequestSecurityTicketForThread(False)


@middleware
def WaitForResponseMiddleware(app, wsgi_env, start_response):
  """WSGI middleware wrapper that waits until response is ready.

  Some middlewares here and some external middlewares rely on behavior that app
  calls finish when response is ready and they do post-processing that should
  be done when handler has returned all response.
  This middleware added for compatibility with such middlewares,
  once all users migrated to Python3, all middlewares should use `yield app()`
  instead of `return app()` and this one should be removed.

  Args:
    app: (callable) a WSGI app per PEP 3333.
    wsgi_env: see PEP 3333
    start_response: see PEP 3333

  Returns:
    A wrapped <app>, which is also a valid WSGI app.
  """
  return list(app(wsgi_env, start_response))


@middleware
def ErrorLoggingMiddleware(app, wsgi_env, start_response):
  """Catch and log unhandled errors for the given app."""
  try:
    return app(wsgi_env, start_response)
  except:








    log_message = traceback.format_exception(
        sys.exc_info()[0],
        sys.exc_info()[1],
        sys.exc_info()[2])

    logging.error(''.join(log_message))
    raise


def _MakeRequestIdHash(log_id):
  """Invent a REQUEST_ID_HASH to replace the one the python27g runtime had."""









  request_id_hash = binascii.crc32(six.ensure_binary(log_id))
  request_id_hash &= 0xffffffff
  request_id_hash_hex = '%08X' % request_id_hash
  return request_id_hash_hex


@middleware
def WsgiEnvSettingMiddleware(app, wsgi_env, start_response):
  """Initialize wsgi_env with reasonable values derived from HTTP headers."""

  https = wsgi_env.get('HTTP_X_APPENGINE_HTTPS')
  if https is not None:
    wsgi_env['HTTPS'] = https










  protocol = wsgi_env.get('HTTP_X_FORWARDED_PROTO')
  if protocol is not None:
    wsgi_env['wsgi.url_scheme'] = protocol
    if protocol == 'http':
      wsgi_env['SERVER_PORT'] = '80'
    elif protocol == 'https':
      wsgi_env['SERVER_PORT'] = '443'
    else:
      logging.warning('Unrecognized value for HTTP_X_FORWARDED_PROTO (%s)'
                      ", won't modify SERVER_PORT", protocol)

  http_host = wsgi_env.get('HTTP_HOST')
  if http_host is not None:
    server_name = urllib.parse.urlparse('//' + http_host).hostname
    wsgi_env['SERVER_NAME'] = server_name



  user_ip = wsgi_env.get('HTTP_X_APPENGINE_USER_IP')
  if user_ip is not None:
    wsgi_env['REMOTE_ADDR'] = user_ip

  return app(wsgi_env, start_response)


@middleware
def SetContextFromHeadersMiddleware(app, wsgi_env, start_response):
  """Set the contextvars from the X_APPENGINE headers."""
  context.init_from_wsgi_environ(wsgi_env)
  return app(wsgi_env, start_response)


@middleware
def OverrideHttpHeadersFromOsEnvironMiddleware(app, wsgi_env, start_response):
  """Override certain HTTP headers with env vars for testing."""
  for key in LOCAL_OVERRIDABLE_VARS:
    val = os.environ.get(key)
    if val is not None:
      wsgi_env['HTTP_X_APPENGINE_'+key] = val
  return app(wsgi_env, start_response)


@middleware
def LegacyWsgiRemoveXAppenginePrefixMiddleware(app, wsgi_env, start_response):
  """Reset HTTP_X_APPENGINE_<foo> as just <foo>."""

  wsgi_env.setdefault('HTTP_X_APPENGINE_AUTH_DOMAIN', 'gmail.com')
  wsgi_env.setdefault('HTTP_X_APPENGINE_USER_IS_ADMIN', '0')
  for key in ENV_VARS_FROM_HTTP_X_APPENGINE_HEADERS:
    value = wsgi_env.get('HTTP_X_APPENGINE_' + key)
    if value is not None:
      wsgi_env[key] = value
  return app(wsgi_env, start_response)


def MakeLegacyWsgiEnvSettingMiddleware(threadsafe=None):
  """Set wsgi_env like it was set in python27."""

  @middleware
  def LegacyWsgiEnvSettingMiddleware(app, wsgi_env, start_response):



    wsgi_env['CLOUD_TRACE_ENABLE_STACK_TRACE'] = ''
    if 'GAE_RUNTIME' in os.environ:
      wsgi_env['GAE_RUNTIME'] = os.environ['GAE_RUNTIME']
    if threadsafe is not None:
      wsgi_env['wsgi.multithread'] = threadsafe
    log_id = wsgi_env.get('HTTP_X_APPENGINE_REQUEST_LOG_ID')
    if log_id is not None:
      wsgi_env['HTTP_X_APPENGINE_REQUEST_ID_HASH'] = _MakeRequestIdHash(log_id)
    return app(wsgi_env, start_response)

  return LegacyWsgiEnvSettingMiddleware


def MakeInitLegacyRequestOsEnvironMiddleware():
  """Patch os.environ to be thread local, and stamp it with default values.

  When this function is called, we remember the values of os.environ. When the
  wrapped inner function (i.e. the WSGI middleware) is called, we patch
  os.environ to be thread local, and we fill in the remembered values, and merge
  in WSGI env vars.

  Returns:
    The InitLegacyRequestOsEnviron Middleware.
  """




  original_environ = dict(os.environ)


  request_environment.PatchOsEnviron()


  os.environ.update(original_environ)

  @middleware
  def InitLegacyRequestOsEnvironMiddleware(app, wsgi_env, start_response):
    """The middleware WSGI app."""







    request_environment.current_request.Init(
        errors=None,
        environ=original_environ.copy())

    return app(wsgi_env, start_response)

  return InitLegacyRequestOsEnvironMiddleware


@middleware
def LegacyCopyWsgiEnvToOsEnvMiddleware(app, wsgi_env, start_response):
  """Copy WSGI variables to os.environ.

  When the wrapped inner function (i.e. the WSGI middleware) is called, we copy
  WSGI environment variables to os.environ. As a precondition, we assume
  os.environ has already been patched to be request-specific.

  Args:
    app: The WSGI app to wrap.
    wsgi_env: see PEP 3333
    start_response: see PEP 3333

  Returns:
    The wrapped app, also a WSGI app.
  """

  assert isinstance(os.environ, request_environment.RequestLocalEnviron)
  for key, val in six.iteritems(wsgi_env):
    if isinstance(val, six.string_types):
      os.environ[key] = val
  return app(wsgi_env, start_response)


@middleware
def CallbackMiddleware(app, wsgi_env, start_response):
  """Calls the request-end callback that the app may have set."""
  try:
    return app(wsgi_env, start_response)
  finally:
    callback.InvokeCallbacks()


@middleware
def RunInNewContextMiddleware(app, wsgi_env, start_response):
  """Runs the app in a copy of the current context.

  Start every request by calling contextvars.copy_context() and running the
  rest of the request in the newly copied context. This ensures that
  each request has its own context that is isolated from all other requests.

  In other words, this "scopes" all ContextVars to the current request.

  Args:
    app: the WSGI app to wrap
    wsgi_env: see PEP 3333
    start_response: see PEP 3333
  Returns:
    The wrapped WSGI app
  """
  ctx = contextvars.copy_context()
  return ctx.run(app, wsgi_env, start_response)



WARMUP_IP = '0.1.0.3'


@middleware
def BackgroundAndShutdownMiddleware(app, wsgi_env, start_response):
  path = wsgi_env['PATH_INFO']
  if (path == '/_ah/background' and wsgi_env.get('REMOTE_ADDR') == WARMUP_IP):
    return background.App(wsgi_env, start_response)
  elif (path == '/_ah/stop' and wsgi_env.get('REMOTE_ADDR') == WARMUP_IP):
    runtime.__BeginShutdown()
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return [b'ok']
  return app(wsgi_env, start_response)


@middleware
def AddDeferredMiddleware(app, wsgi_env, start_response):
  """Intercept calls to the default endpoint for Deferred.

  Handle requests made to /_ah/queue/deferred

  Args:
    app: the WSGI app to wrap
    wsgi_env: see PEP 3333
    start_response: see PEP 3333
  Returns:
    The wrapped WSGI app response in UTF-8 format
  """
  path = wsgi_env['PATH_INFO']
  if path == '/_ah/queue/deferred':
    return deferred.application(wsgi_env, start_response)
  return app(wsgi_env, start_response)


@middleware
def SetNamespaceFromHeader(app, wsgi_env, start_response):
  ns_from_header = wsgi_env.get('HTTP_X_APPENGINE_CURRENT_NAMESPACE')
  if ns_from_header is not None:
    namespace_manager.set_namespace(ns_from_header)
  return app(wsgi_env, start_response)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#














"""A tasklet decorator.

Tasklets are a way to write concurrently running functions without
threads; tasklets are executed by an event loop and can suspend
themselves blocking for I/O or some other operation using a yield
statement.  The notion of a blocking operation is abstracted into the
Future class, but a tasklet may also yield an RPC in order to wait for
that RPC to complete.

The @tasklet decorator wraps generator function so that when it is
called, a Future is returned while the generator is executed by the
event loop.  Within the tasklet, any yield of a Future waits for and
returns the Future's result.  For example::

  @tasklet
  def foo():
    a = yield <some Future>
    b = yield <another Future>
    raise Return(a + b)

  def main():
    f = foo()
    x = f.get_result()
    print x

Note that blocking until the Future's result is available using
get_result() is somewhat inefficient (though not vastly -- it is not
busy-waiting).  In most cases such code should be rewritten as a tasklet
instead::

  @tasklet
  def main_tasklet():
    f = foo()
    x = yield f
    print x

Calling a tasklet automatically schedules it with the event loop::

  def main():
    f = main_tasklet()
    eventloop.run()  # Run until no tasklets left to do
    f.done()  # Returns True

As a special feature, if the wrapped function is not a generator
function, its return value is returned via the Future.  This makes the
following two equivalent::

  @tasklet
  def foo():
    return 42

  @tasklet
  def foo():
    if False: yield  # The presence of 'yield' makes foo a generator
    raise Return(42)  # Or, after PEP 380, return 42

This feature (inspired by Monocle) is handy in case you are
implementing an interface that expects tasklets but you have no need to
suspend -- there's no need to insert a dummy yield in order to make
the tasklet into a generator.
"""

import collections
import functools
import logging
import os
import sys
import threading
import types
import weakref

from google.appengine.api import apiproxy_rpc
from google.appengine.api import apiproxy_stub
from google.appengine.api import apiproxy_stub_map
from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import full_app_id
from google.appengine.api import namespace_manager
from google.appengine.datastore import datastore_pbs
from google.appengine.datastore import datastore_rpc
from google.appengine.ext.ndb import eventloop
from google.appengine.ext.ndb import utils
from google.appengine.runtime import apiproxy
import six
from six.moves import map

__all__ = ['Return', 'tasklet', 'synctasklet', 'toplevel', 'sleep',
           'add_flow_exception', 'get_return_value',
           'get_context', 'set_context',
           'make_default_context', 'make_context',
           'Future', 'MultiFuture', 'QueueFuture', 'SerialQueueFuture',
           'ReducingFuture',
          ]

_logging_debug = utils.logging_debug

_CALLBACK_KEY = '__CALLBACK__'


def _is_generator(obj):
  """Helper to test for a generator object.

  NOTE: This tests for the (iterable) object returned by calling a
  generator function, not for a generator function.
  """
  return isinstance(obj, types.GeneratorType)


class _State(threading.local):
  """Hold thread-local state."""

  def __init__(self):
    super(_State, self).__init__()
    self.current_context = None
    self.all_generators = weakref.WeakSet()
    self.all_pending = set()

  def set_context(self, ctx):
    self.current_context = ctx

  def add_generator(self, gen):
    if _CALLBACK_KEY not in os.environ:
      apiproxy.SetRequestEndCallback(self.reset)
      os.environ[_CALLBACK_KEY] = '1'

    _logging_debug('all_generators: add %s', gen)
    self.all_generators.add(gen)

  def add_pending(self, fut):
    if _CALLBACK_KEY not in os.environ:
      apiproxy.SetRequestEndCallback(self.reset)
      os.environ[_CALLBACK_KEY] = '1'

    _logging_debug('all_pending: add %s', fut)
    self.all_pending.add(fut)

  def remove_pending(self, fut, status='success'):
    if fut in self.all_pending:
      _logging_debug('all_pending: %s: remove %s', status, fut)
      self.all_pending.remove(fut)
    else:
      _logging_debug('all_pending: %s: not found %s', status, fut)

  def clear_all_generators(self):
    if self.all_generators:
      _logging_debug('all_generators: clear %s', self.all_generators)
      for gen in self.all_generators:
        gen.close()
      self.all_generators.clear()
    else:
      _logging_debug('all_generators: clear no-op')

  def clear_all_pending(self):
    if self.all_pending:
      _logging_debug('all_pending: clear %s', self.all_pending)
      self.all_pending.clear()
    else:
      _logging_debug('all_pending: clear no-op')

  def dump_all_pending(self, verbose=False):
    pending = []
    for fut in self.all_pending:
      if verbose:
        line = fut.dump() + ('\n' + '-' * 40)
      else:
        line = fut.dump_stack()
      pending.append(line)
    return '\n'.join(pending)

  def reset(self, unused_req_id):
    self.current_context = None
    ev = eventloop.get_event_loop()
    ev.clear()
    self.clear_all_pending()
    self.clear_all_generators()


_state = _State()



_flow_exceptions = ()


def add_flow_exception(exc):
  """Add an exception that should not be logged.

  The argument must be a subclass of Exception.
  """
  global _flow_exceptions
  if not isinstance(exc, type) or not issubclass(exc, Exception):
    raise TypeError('Expected an Exception subclass, got %r' % (exc,))
  as_set = set(_flow_exceptions)
  as_set.add(exc)
  _flow_exceptions = tuple(as_set)


def _init_flow_exceptions():
  """Internal helper to initialize _flow_exceptions.

  This automatically adds webob.exc.HTTPException, if it can be imported.
  """
  global _flow_exceptions
  _flow_exceptions = ()
  add_flow_exception(datastore_errors.Rollback)
  try:
    from webob import exc
  except ImportError:
    pass
  else:
    add_flow_exception(exc.HTTPException)


_init_flow_exceptions()


class Future(object):
  """A Future has 0 or more callbacks.

  The callbacks will be called when the result is ready.

  NOTE: This is somewhat inspired but not conformant to the Future interface
  defined by PEP 3148.  It is also inspired (and tries to be somewhat
  compatible with) the App Engine specific UserRPC and MultiRpc classes.
  """




  IDLE = apiproxy_rpc.RPC.IDLE
  RUNNING = apiproxy_rpc.RPC.RUNNING
  FINISHING = apiproxy_rpc.RPC.FINISHING



  _geninfo = None

  def __init__(self, info=None):


    __ndb_debug__ = 'SKIP'
    self._info = info
    self._where = utils.get_stack()
    self._context = None
    self._reset()

  def _reset(self):
    self._done = False
    self._result = None
    self._exception = None
    self._traceback = None
    self._callbacks = []
    self._immediate_callbacks = []
    _state.add_pending(self)
    self._next = None




  def __repr__(self):
    if self._done:
      if self._exception is not None:
        state = 'exception %s: %s' % (self._exception.__class__.__name__,
                                      self._exception)
      else:
        state = 'result %r' % (self._result,)
    else:
      state = 'pending'
    line = '?'
    for line in self._where:
      if 'tasklets.py' not in line:
        break
    if self._info:
      line += ' for %s' % self._info
    if self._geninfo:
      line += ' %s' % self._geninfo
    return '<%s %x created by %s; %s>' % (
        self.__class__.__name__, id(self), line, state)

  def dump(self):
    return '%s\nCreated by %s' % (self.dump_stack(),
                                  '\n called by '.join(self._where))

  def dump_stack(self):
    lines = []
    fut = self
    while fut is not None:
      lines.append(str(fut))
      fut = fut._next
    return '\n waiting for '.join(lines)

  def add_callback(self, callback, *args, **kwds):
    if self._done:
      eventloop.queue_call(None, callback, *args, **kwds)
    else:
      self._callbacks.append((callback, args, kwds))

  def add_immediate_callback(self, callback, *args, **kwds):
    if self._done:
      callback(*args, **kwds)
    else:
      self._immediate_callbacks.append((callback, args, kwds))

  def set_result(self, result):
    if self._done:
      raise RuntimeError('Result cannot be set twice.')
    self._result = result
    self._done = True
    _state.remove_pending(self)
    for callback, args, kwds in self._immediate_callbacks:
      callback(*args, **kwds)
    for callback, args, kwds in self._callbacks:
      eventloop.queue_call(None, callback, *args, **kwds)

  def set_exception(self, exc, tb=None):
    if not isinstance(exc, BaseException):
      raise TypeError('exc must be an Exception; received %r' % exc)
    if self._done:
      raise RuntimeError('Exception cannot be set twice.')
    self._exception = exc
    self._traceback = tb
    self._done = True
    _state.remove_pending(self, status='fail')
    for callback, args, kwds in self._immediate_callbacks:
      callback(*args, **kwds)
    for callback, args, kwds in self._callbacks:
      eventloop.queue_call(None, callback, *args, **kwds)

  def done(self):
    return self._done

  @property
  def state(self):


    if self._done:
      return self.FINISHING
    else:
      return self.RUNNING

  def wait(self):
    if self._done:
      return
    ev = eventloop.get_event_loop()
    while not self._done:
      if not ev.run1():
        logging.info('Deadlock in %s', self)
        logging.info('All pending Futures:\n%s', _state.dump_all_pending())
        _logging_debug('All pending Futures (verbose):\n%s',
                       _state.dump_all_pending(verbose=True))
        self.set_exception(RuntimeError('Deadlock waiting for %s' % self))

  def get_exception(self):
    self.wait()
    return self._exception

  def get_traceback(self):
    self.wait()
    return self._traceback

  def check_success(self):
    self.wait()
    if self._exception is not None:
      six.reraise(self._exception.__class__, self._exception, self._traceback)

  def get_result(self):
    self.check_success()
    return self._result


  @classmethod
  def wait_any(cls, futures):

    waiting_on = set(futures)
    ev = eventloop.get_event_loop()
    while waiting_on:
      for f in waiting_on:
        if f.state == cls.FINISHING:
          return f
      ev.run1()
    return None


  @classmethod
  def wait_all(cls, futures):

    waiting_on = set(futures)
    ev = eventloop.get_event_loop()
    while waiting_on:
      waiting_on = set(f for f in waiting_on if f.state == cls.RUNNING)
      ev.run1()

  def _help_tasklet_along(self, ns, ds_conn, gen, val=None, exc=None, tb=None):

    info = utils.gen_info(gen)

    __ndb_debug__ = info
    try:
      save_context = get_context()
      save_namespace = namespace_manager.get_namespace()
      save_ds_connection = datastore._GetConnection()
      try:
        set_context(self._context)
        if ns != save_namespace:
          namespace_manager.set_namespace(ns)
        if ds_conn is not save_ds_connection:
          datastore._SetConnection(ds_conn)
        if exc is not None:
          _logging_debug('Throwing %s(%s) into %s',
                         exc.__class__.__name__, exc, info)
          value = gen.throw(exc.__class__, exc, tb)
        else:
          _logging_debug('Sending %r to %s', val, info)
          value = gen.send(val)
          self._context = get_context()
      finally:
        ns = namespace_manager.get_namespace()
        ds_conn = datastore._GetConnection()
        set_context(save_context)
        if save_namespace != ns:
          namespace_manager.set_namespace(save_namespace)
        if save_ds_connection is not ds_conn:
          datastore._SetConnection(save_ds_connection)

    except (StopIteration, Return) as err:
      result = get_return_value(err)
      _logging_debug('%s returned %r', info, result)
      self.set_result(result)
      return

    except GeneratorExit:




      raise

    except Exception as err:
      _, _, tb = sys.exc_info()
      if isinstance(err, _flow_exceptions):


        _logging_debug('%s raised %s(%s)',
                       info, err.__class__.__name__, err)
      elif utils.DEBUG and logging.getLogger().level < logging.DEBUG:


        logging.warning('%s raised %s(%s)',
                        info, err.__class__.__name__, err, exc_info=True)
      else:

        logging.warning('%s raised %s(%s)', info, err.__class__.__name__, err)
      self.set_exception(err, tb)
      return

    else:
      _logging_debug('%s yielded %r', info, value)
      if isinstance(value, (apiproxy_stub_map.UserRPC,
                            datastore_rpc.MultiRpc)):

        eventloop.queue_rpc(value, self._on_rpc_completion,
                            value, ns, ds_conn, gen)
        return
      if isinstance(value, Future):

        if self._next:
          raise RuntimeError('Future has already completed yet next is %r' %
                             self._next)
        self._next = value
        self._geninfo = utils.gen_info(gen)
        _logging_debug('%s is now blocked waiting for %s', self, value)
        value.add_callback(self._on_future_completion, value, ns, ds_conn, gen)
        return
      if isinstance(value, (tuple, list)):

        info = 'multi-yield from %s' % utils.gen_info(gen)
        mfut = MultiFuture(info)
        try:
          for subfuture in value:
            mfut.add_dependent(subfuture)
          mfut.complete()
        except GeneratorExit:
          raise
        except Exception as err:
          _, _, tb = sys.exc_info()
          mfut.set_exception(err, tb)
        mfut.add_callback(self._on_future_completion, mfut, ns, ds_conn, gen)
        return
      if _is_generator(value):

        raise NotImplementedError('Cannot defer to another generator.')
      raise RuntimeError('A tasklet should not yield a plain value: '
                         '%.200s yielded %.200r' % (info, value))

  def _on_rpc_completion(self, rpc, ns, ds_conn, gen):
    try:
      result = rpc.get_result()
    except GeneratorExit:
      raise
    except Exception as err:
      _, _, tb = sys.exc_info()
      self._help_tasklet_along(ns, ds_conn, gen, exc=err, tb=tb)
    else:
      self._help_tasklet_along(ns, ds_conn, gen, result)

  def _on_future_completion(self, future, ns, ds_conn, gen):
    if self._next is future:
      self._next = None
      self._geninfo = None
      _logging_debug('%s is no longer blocked waiting for %s', self, future)
    exc = future.get_exception()
    if exc is not None:
      self._help_tasklet_along(ns, ds_conn, gen,
                               exc=exc, tb=future.get_traceback())
    else:
      val = future.get_result()
      self._help_tasklet_along(ns, ds_conn, gen, val)


def sleep(dt):
  """Public function to sleep some time.

  Example:
    yield tasklets.sleep(0.5)  # Sleep for half a sec.
  """
  fut = Future('sleep(%.3f)' % dt)
  eventloop.queue_call(dt, fut.set_result, None)
  return fut


class MultiFuture(Future):
  """A Future that depends on multiple other Futures.

  This is used internally by 'v1, v2, ... = yield f1, f2, ...'; the
  semantics (e.g. error handling) are constrained by that use case.

  The protocol from the caller's POV is::

    mf = MultiFuture()
    mf.add_dependent(<some other Future>)  -OR- mf.putq(<some value>)
    mf.add_dependent(<some other Future>)  -OR- mf.putq(<some value>)
      .
      . (More mf.add_dependent() and/or mf.putq() calls)
      .
    mf.complete()  # No more dependents will be added.
      .
      . (Time passes)
      .
    results = mf.get_result()

  Now, results is a list of results from all dependent Futures in
  the order in which they were added.

  It is legal to add the same dependent multiple times.

  Callbacks can be added at any point.

  From a dependent Future POV, there's nothing to be done: a callback
  is automatically added to each dependent Future which will signal
  its completion to the MultiFuture.

  Error handling: if any dependent future raises an error, it is
  propagated to mf.  To force an early error, you can call
  mf.set_exception() instead of mf.complete().  After this you can't
  call mf.add_dependent() or mf.putq() any more.
  """

  def __init__(self, info=None):

    __ndb_debug__ = 'SKIP'
    self._full = False
    self._dependents = set()
    self._results = []
    super(MultiFuture, self).__init__(info=info)

  def __repr__(self):


    line = super(MultiFuture, self).__repr__()
    lines = [line]
    for fut in self._results:
      lines.append(fut.dump_stack().replace('\n', '\n  '))
    return '\n waiting for '.join(lines)



  def complete(self):
    if self._full:
      raise RuntimeError('MultiFuture cannot complete twice.')
    self._full = True
    if not self._dependents:
      self._finish()


  def set_exception(self, exc, tb=None):
    self._full = True
    super(MultiFuture, self).set_exception(exc, tb)

  def _finish(self):
    if not self._full:
      raise RuntimeError('MultiFuture cannot finish until completed.')
    if self._dependents:
      raise RuntimeError('MultiFuture cannot finish whilst waiting for '
                         'dependents %r' % self._dependents)
    if self._done:
      raise RuntimeError('MultiFuture done before finishing.')
    try:
      result = [r.get_result() for r in self._results]
    except GeneratorExit:
      raise
    except Exception as err:
      _, _, tb = sys.exc_info()
      self.set_exception(err, tb)
    else:
      self.set_result(result)

  def putq(self, value):
    if isinstance(value, Future):
      fut = value
    else:
      fut = Future()
      fut.set_result(value)
    self.add_dependent(fut)

  def add_dependent(self, fut):
    if isinstance(fut, list):
      mfut = MultiFuture()
      list(map(mfut.add_dependent, fut))
      mfut.complete()
      fut = mfut
    elif not isinstance(fut, Future):
      raise TypeError('Expected Future, received %s: %r' % (type(fut), fut))
    if self._full:
      raise RuntimeError('MultiFuture cannot add a dependent once complete.')
    self._results.append(fut)
    if fut not in self._dependents:
      self._dependents.add(fut)
      fut.add_callback(self._signal_dependent_done, fut)

  def _signal_dependent_done(self, fut):
    self._dependents.remove(fut)
    if self._full and not self._dependents and not self._done:
      self._finish()


class QueueFuture(Future):
  """A Queue following the same protocol as MultiFuture.

  However, instead of returning results as a list, it lets you
  retrieve results as soon as they are ready, one at a time, using
  getq().  The Future itself finishes with a result of None when the
  last result is ready (regardless of whether it was retrieved).

  The getq() method returns a Future which blocks until the next
  result is ready, and then returns that result.  Each getq() call
  retrieves one unique result.  Extra getq() calls after the last
  result is already returned return EOFError as their Future's
  exception.  (I.e., q.getq() returns a Future as always, but yieding
  that Future raises EOFError.)

  NOTE: Values can also be pushed directly via .putq(value).  However
  there is no flow control -- if the producer is faster than the
  consumer, the queue will grow unbounded.
  """


  def __init__(self, info=None):
    self._full = False
    self._dependents = set()
    self._completed = collections.deque()
    self._waiting = collections.deque()


    super(QueueFuture, self).__init__(info=info)



  def complete(self):
    if self._full:
      raise RuntimeError('MultiFuture cannot complete twice.')
    self._full = True
    if not self._dependents:
      self.set_result(None)
      self._mark_finished()

  def set_exception(self, exc, tb=None):
    self._full = True
    super(QueueFuture, self).set_exception(exc, tb)
    if not self._dependents:
      self._mark_finished()

  def putq(self, value):
    if isinstance(value, Future):
      fut = value
    else:
      fut = Future()
      fut.set_result(value)
    self.add_dependent(fut)

  def add_dependent(self, fut):
    if not isinstance(fut, Future):
      raise TypeError('fut must be a Future instance; received %r' % fut)
    if self._full:
      raise RuntimeError('QueueFuture add dependent once complete.')
    if fut not in self._dependents:
      self._dependents.add(fut)
      fut.add_callback(self._signal_dependent_done, fut)

  def _signal_dependent_done(self, fut):
    if not fut.done():
      raise RuntimeError('Future not done before signalling dependant done.')
    self._dependents.remove(fut)
    exc = fut.get_exception()
    tb = fut.get_traceback()
    val = None
    if exc is None:
      val = fut.get_result()
    if self._waiting:
      waiter = self._waiting.popleft()
      self._pass_result(waiter, exc, tb, val)
    else:
      self._completed.append((exc, tb, val))
    if self._full and not self._dependents and not self._done:
      self.set_result(None)
      self._mark_finished()

  def _mark_finished(self):
    if not self.done():
      raise RuntimeError('Future not done before marking as finished.')
    while self._waiting:
      waiter = self._waiting.popleft()
      self._pass_eof(waiter)

  def getq(self):
    fut = Future()
    if self._completed:
      exc, tb, val = self._completed.popleft()
      self._pass_result(fut, exc, tb, val)
    elif self._full and not self._dependents:
      self._pass_eof(fut)
    else:
      self._waiting.append(fut)
    return fut

  def _pass_eof(self, fut):
    if not self._done:
      raise RuntimeError('QueueFuture cannot pass EOF until done.')
    exc = self.get_exception()
    if exc is not None:
      tb = self.get_traceback()
    else:
      exc = EOFError('Queue is empty')
      tb = None
    self._pass_result(fut, exc, tb, None)

  def _pass_result(self, fut, exc, tb, val):
    if exc is not None:
      fut.set_exception(exc, tb)
    else:
      fut.set_result(val)


class SerialQueueFuture(Future):
  """Like QueueFuture but maintains the order of insertion.

  This class is used by Query operations.

  Invariants:

  - At least one of _queue and _waiting is empty.
  - The Futures in _waiting are always pending.

  (The Futures in _queue may be pending or completed.)

  In the discussion below, add_dependent() is treated the same way as
  putq().

  If putq() is ahead of getq(), the situation is like this:

                         putq()
                         v
    _queue: [f1, f2, ...]; _waiting: []
    ^
    getq()

  Here, putq() appends a Future to the right of _queue, and getq()
  removes one from the left.

  If getq() is ahead of putq(), it's like this:

              putq()
              v
    _queue: []; _waiting: [f1, f2, ...]
                                       ^
                                       getq()

  Here, putq() removes a Future from the left of _waiting, and getq()
  appends one to the right.

  When both are empty, putq() appends a Future to the right of _queue,
  while getq() appends one to the right of _waiting.

  The _full flag means that no more calls to putq() will be made; it
  is set by calling either complete() or set_exception().

  Calling complete() signals that no more putq() calls will be made.
  If getq() is behind, subsequent getq() calls will eat up _queue
  until it is empty, and after that will return a Future that passes
  EOFError (note that getq() itself never raises EOFError).  If getq()
  is ahead when complete() is called, the Futures in _waiting are all
  passed an EOFError exception (thereby eating up _waiting).

  If, instead of complete(), set_exception() is called, the exception
  and traceback set there will be used instead of EOFError.
  """

  def __init__(self, info=None):
    self._queue = collections.deque()
    self._waiting = collections.deque()
    super(SerialQueueFuture, self).__init__(info=info)



  def complete(self):
    while self._waiting:
      waiter = self._waiting.popleft()
      waiter.set_exception(EOFError('Queue is empty'))



    self.set_result(None)

  def set_exception(self, exc, tb=None):
    super(SerialQueueFuture, self).set_exception(exc, tb)
    while self._waiting:
      waiter = self._waiting.popleft()
      waiter.set_exception(exc, tb)

  def putq(self, value):
    if isinstance(value, Future):
      fut = value
    else:
      if self._waiting:
        waiter = self._waiting.popleft()
        waiter.set_result(value)
        return
      fut = Future()
      fut.set_result(value)
    self.add_dependent(fut)

  def add_dependent(self, fut):
    if not isinstance(fut, Future):
      raise TypeError('fut must be a Future instance; received %r' % fut)
    if self._done:
      raise RuntimeError('SerialQueueFuture cannot add dependent '
                         'once complete.')
    if self._waiting:
      waiter = self._waiting.popleft()
      fut.add_callback(_transfer_result, fut, waiter)
    else:
      self._queue.append(fut)

  def getq(self):
    if self._queue:
      fut = self._queue.popleft()
    else:
      fut = Future()
      if self._done:
        err = self.get_exception()
        if err is not None:
          tb = self.get_traceback()
        else:
          err = EOFError('Queue is empty')
          tb = None
        fut.set_exception(err, tb)
      else:
        self._waiting.append(fut)
    return fut


def _transfer_result(fut1, fut2):
  """Helper to transfer result or errors from one Future to another."""
  exc = fut1.get_exception()
  if exc is not None:
    tb = fut1.get_traceback()
    fut2.set_exception(exc, tb)
  else:
    val = fut1.get_result()
    fut2.set_result(val)


class ReducingFuture(Future):
  """A Queue following the same protocol as MultiFuture.

  However the result, instead of being a list of results of dependent
  Futures, is computed by calling a 'reducer' tasklet.  The reducer tasklet
  takes a list of values and returns a single value.  It may be called
  multiple times on sublists of values and should behave like
  e.g. sum().

  NOTE: The reducer input values may be reordered compared to the
  order in which they were added to the queue.
  """


  def __init__(self, reducer, info=None, batch_size=20):
    self._reducer = reducer
    self._batch_size = batch_size
    self._full = False
    self._dependents = set()
    self._completed = collections.deque()
    self._queue = collections.deque()
    super(ReducingFuture, self).__init__(info=info)



  def complete(self):
    if self._full:
      raise RuntimeError('ReducingFuture cannot complete twice.')
    self._full = True
    if not self._dependents:
      self._mark_finished()

  def set_exception(self, exc, tb=None):
    self._full = True
    self._queue.clear()
    super(ReducingFuture, self).set_exception(exc, tb)

  def putq(self, value):
    if isinstance(value, Future):
      fut = value
    else:
      fut = Future()
      fut.set_result(value)
    self.add_dependent(fut)

  def add_dependent(self, fut):
    if self._full:
      raise RuntimeError('ReducingFuture cannot add dependent once complete.')
    self._internal_add_dependent(fut)

  def _internal_add_dependent(self, fut):
    if not isinstance(fut, Future):
      raise TypeError('fut must be a Future; received %r' % fut)
    if fut not in self._dependents:
      self._dependents.add(fut)
      fut.add_callback(self._signal_dependent_done, fut)

  def _signal_dependent_done(self, fut):
    if not fut.done():
      raise RuntimeError('Future not done before signalling dependant done.')
    self._dependents.remove(fut)
    if self._done:
      return
    try:
      val = fut.get_result()
    except GeneratorExit:
      raise
    except Exception as err:
      _, _, tb = sys.exc_info()
      self.set_exception(err, tb)
      return
    self._queue.append(val)
    if len(self._queue) >= self._batch_size:
      todo = list(self._queue)
      self._queue.clear()
      try:
        nval = self._reducer(todo)
      except GeneratorExit:
        raise
      except Exception as err:
        _, _, tb = sys.exc_info()
        self.set_exception(err, tb)
        return
      if isinstance(nval, Future):
        self._internal_add_dependent(nval)
      else:
        self._queue.append(nval)
    if self._full and not self._dependents:
      self._mark_finished()

  def _mark_finished(self):
    if not self._queue:
      self.set_result(None)
    elif len(self._queue) == 1:
      self.set_result(self._queue.pop())
    else:
      todo = list(self._queue)
      self._queue.clear()
      try:
        nval = self._reducer(todo)
      except GeneratorExit:
        raise
      except Exception as err:
        _, _, tb = sys.exc_info()
        self.set_exception(err, tb)
        return
      if isinstance(nval, Future):
        self._internal_add_dependent(nval)
      else:
        self.set_result(nval)


class Return(Exception):
  """Return from a tasklet in Python 2.

  In Python 2, generators may not return a value. In order to return a value
  from a tasklet, then, it is necessary to raise an instance of this
  exception with the return value:

      @ndb.tasklet
      def get_some_stuff():
        future1 = get_something_async()
        future2 = get_something_else_async()
        thing1, thing2 = yield future1, future2
        result = compute_result(thing1, thing2)
        raise ndb.Return(result)

  In Python 3, you can simply return the result:

      @ndb.tasklet
      def get_some_stuff():
        future1 = get_something_async()
        future2 = get_something_else_async()
        thing1, thing2 = yield future1, future2
        result = compute_result(thing1, thing2)
        return result
  """


def get_return_value(err):

  if not err.args:
    result = None
  elif len(err.args) == 1:
    result = err.args[0]
  else:
    result = err.args
  return result


def tasklet(func):


  @functools.wraps(func)
  def tasklet_wrapper(*args, **kwds):






    __ndb_debug__ = utils.func_info(func)
    fut = Future('tasklet %s' % utils.func_info(func))
    fut._context = get_context()
    try:
      result = func(*args, **kwds)
    except (StopIteration, Return) as err:


      result = get_return_value(err)
    if _is_generator(result):
      ns = namespace_manager.get_namespace()
      ds_conn = datastore._GetConnection()
      _state.add_generator(result)
      eventloop.queue_call(None, fut._help_tasklet_along, ns, ds_conn, result)
    else:
      fut.set_result(result)
    return fut

  return tasklet_wrapper


def synctasklet(func):
  """Decorator to run a function as a tasklet when called.

  Use this to wrap a request handler function that will be called by
  some web application framework (e.g. a Django view function or a
  webapp.RequestHandler.get method).
  """
  taskletfunc = tasklet(func)

  @functools.wraps(func)
  def synctasklet_wrapper(*args, **kwds):

    __ndb_debug__ = utils.func_info(func)
    return taskletfunc(*args, **kwds).get_result()
  return synctasklet_wrapper


def toplevel(func):
  """A sync tasklet that sets a fresh default Context.

  Use this for toplevel view functions such as
  webapp.RequestHandler.get() or Django view functions.
  """
  synctaskletfunc = synctasklet(func)

  @functools.wraps(func)
  def add_context_wrapper(*args, **kwds):

    __ndb_debug__ = utils.func_info(func)
    _state.clear_all_pending()

    ctx = make_default_context()
    try:
      set_context(ctx)
      return synctaskletfunc(*args, **kwds)
    finally:
      set_context(None)
      ctx.flush().check_success()
      eventloop.run()
  return add_context_wrapper


_CONTEXT_KEY = '__CONTEXT__'

_DATASTORE_APP_ID_ENV = 'DATASTORE_APP_ID'
_DATASTORE_PROJECT_ID_ENV = 'DATASTORE_PROJECT_ID'
_DATASTORE_ADDITIONAL_APP_IDS_ENV = 'DATASTORE_ADDITIONAL_APP_IDS'
_DATASTORE_USE_PROJECT_ID_AS_APP_ID_ENV = 'DATASTORE_USE_PROJECT_ID_AS_APP_ID'


def get_context():

  ctx = None
  if os.getenv(_CONTEXT_KEY):
    ctx = _state.current_context
  if ctx is None:
    ctx = make_default_context()
    set_context(ctx)
  return ctx


def make_default_context():

  datastore_app_id = os.environ.get(_DATASTORE_APP_ID_ENV, None)
  datastore_project_id = os.environ.get(_DATASTORE_PROJECT_ID_ENV, None)
  if datastore_app_id or datastore_project_id:

    app_id_override = bool(os.environ.get(
        _DATASTORE_USE_PROJECT_ID_AS_APP_ID_ENV, False))
    if not datastore_app_id and not app_id_override:
      raise ValueError('Could not determine app id. To use project id (%s) '
                       'instead, set %s=true. This will affect the '
                       'serialized form of entities and should not be used '
                       'if serialized entities will be shared between '
                       'code running on App Engine and code running off '
                       'App Engine. Alternatively, set %s=<app id>.'
                       % (datastore_project_id,
                          _DATASTORE_USE_PROJECT_ID_AS_APP_ID_ENV,
                          _DATASTORE_APP_ID_ENV))
    elif datastore_app_id:
      if app_id_override:
        raise ValueError('App id was provided (%s) but %s was set to true. '
                         'Please unset either %s or %s.' %
                         (datastore_app_id,
                          _DATASTORE_USE_PROJECT_ID_AS_APP_ID_ENV,
                          _DATASTORE_APP_ID_ENV,
                          _DATASTORE_USE_PROJECT_ID_AS_APP_ID_ENV))
      elif datastore_project_id:

        id_resolver = datastore_pbs.IdResolver([datastore_app_id])
        if (datastore_project_id !=
            id_resolver.resolve_project_id(datastore_app_id)):
          raise ValueError('App id "%s" does not match project id "%s".'
                           % (datastore_app_id, datastore_project_id))

    datastore_app_id = datastore_project_id or datastore_app_id
    additional_app_str = os.environ.get(_DATASTORE_ADDITIONAL_APP_IDS_ENV, '')
    additional_apps = (app.strip() for app in additional_app_str.split(','))
    return _make_cloud_datastore_context(datastore_app_id, additional_apps)
  return make_context()


@utils.positional(0)
def make_context(conn=None, config=None):

  from google.appengine.ext.ndb import context
  return context.Context(conn=conn, config=config)


def _make_cloud_datastore_context(app_id, external_app_ids=()):
  """Creates a new context to connect to a remote Cloud Datastore instance.

  This should only be used outside of Google App Engine.

  Args:
    app_id: The application id to connect to. This differs from the project
      id as it may have an additional prefix, e.g. "s~" or "e~".
    external_app_ids: A list of apps that may be referenced by data in your
      application. For example, if you are connected to s~my-app and store keys
      for s~my-other-app, you should include s~my-other-app in the external_apps
      list.
  Returns:
    An ndb.Context that can connect to a Remote Cloud Datastore. You can use
    this context by passing it to ndb.set_context.
  """
  from google.appengine.ext.ndb import model

  if not datastore_pbs._CLOUD_DATASTORE_ENABLED:
    raise datastore_errors.BadArgumentError(
        datastore_pbs.MISSING_CLOUD_DATASTORE_MESSAGE)
  import googledatastore
  from google.appengine.datastore import cloud_datastore_v1_remote_stub

  current_app_id = full_app_id.get()
  if current_app_id and current_app_id != app_id:


    raise ValueError('Cannot create a Cloud Datastore context that connects '
                     'to an application (%s) that differs from the application '
                     'already connected to (%s).' % (app_id, current_app_id))
  full_app_id.put(app_id)

  id_resolver = datastore_pbs.IdResolver((app_id,) + tuple(external_app_ids))
  project_id = id_resolver.resolve_project_id(app_id)
  endpoint = googledatastore.helper.get_project_endpoint_from_env(project_id)
  datastore = googledatastore.Datastore(
      project_endpoint=endpoint,
      credentials=googledatastore.helper.get_credentials_from_env())

  conn = model.make_connection(_api_version=datastore_rpc._CLOUD_DATASTORE_V1,
                               _id_resolver=id_resolver)


  try:
    stub = cloud_datastore_v1_remote_stub.CloudDatastoreV1RemoteStub(datastore)
    apiproxy_stub_map.apiproxy.RegisterStub(datastore_rpc._CLOUD_DATASTORE_V1,
                                            stub)
  except:
    pass



  try:
    apiproxy_stub_map.apiproxy.RegisterStub('memcache', _ThrowingStub())
  except:
    pass
  try:
    apiproxy_stub_map.apiproxy.RegisterStub('taskqueue', _ThrowingStub())
  except:
    pass


  return make_context(conn=conn)


def set_context(new_context):

  os.environ[_CONTEXT_KEY] = '1'
  _state.set_context(new_context)


class _ThrowingStub(apiproxy_stub.APIProxyStub):
  """A Stub implementation which always throws a NotImplementedError."""

  def __init__(self):
    pass


  def MakeSyncCall(self, service, call, request, response):
    raise NotImplementedError('In order to use %s.%s you must '
                              'install the Remote API.' % (service, call))


  def CreateRPC(self):
    return apiproxy_rpc.RPC(stub=self)

































































#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Image manipulation API.

App Engine provides the ability to manipulate image data using a dedicated
Images service. The Images service can resize, rotate, flip, and crop images; it
can composite multiple images into a single image; and it can convert image data
between several formats. It can also enhance photographs using a predefined
algorithm. The API can also provide information about an image, such as its
format, width, height, and a histogram of color values.
"""








import json
import struct

from google.appengine.api import apiproxy_stub_map
from google.appengine.api import blobstore
from google.appengine.api import datastore_types
from google.appengine.api.images import images_service_pb2
from google.appengine.runtime import apiproxy_errors

import six
from six.moves import range

BlobKey = datastore_types.BlobKey


JPEG = images_service_pb2.OutputSettings.JPEG
PNG = images_service_pb2.OutputSettings.PNG
WEBP = images_service_pb2.OutputSettings.WEBP
BMP = -1
GIF = -2
ICO = -3
TIFF = -4

OUTPUT_ENCODING_TYPES = frozenset([JPEG, PNG, WEBP])

UNCHANGED_ORIENTATION = images_service_pb2.InputSettings.UNCHANGED_ORIENTATION
CORRECT_ORIENTATION = images_service_pb2.InputSettings.CORRECT_ORIENTATION

ORIENTATION_CORRECTION_TYPE = frozenset(
    [UNCHANGED_ORIENTATION, CORRECT_ORIENTATION])

TOP_LEFT = images_service_pb2.CompositeImageOptions.TOP_LEFT
TOP_CENTER = images_service_pb2.CompositeImageOptions.TOP
TOP_RIGHT = images_service_pb2.CompositeImageOptions.TOP_RIGHT
CENTER_LEFT = images_service_pb2.CompositeImageOptions.LEFT
CENTER_CENTER = images_service_pb2.CompositeImageOptions.CENTER
CENTER_RIGHT = images_service_pb2.CompositeImageOptions.RIGHT
BOTTOM_LEFT = images_service_pb2.CompositeImageOptions.BOTTOM_LEFT
BOTTOM_CENTER = images_service_pb2.CompositeImageOptions.BOTTOM
BOTTOM_RIGHT = images_service_pb2.CompositeImageOptions.BOTTOM_RIGHT

ANCHOR_TYPES = frozenset([
    TOP_LEFT, TOP_CENTER, TOP_RIGHT, CENTER_LEFT, CENTER_CENTER, CENTER_RIGHT,
    BOTTOM_LEFT, BOTTOM_CENTER, BOTTOM_RIGHT
])



MAX_TRANSFORMS_PER_REQUEST = 10




MAX_COMPOSITES_PER_REQUEST = 16


class Error(Exception):
  """Base error class for this module."""


class TransformationError(Error):
  """An error occurred while attempting to transform the image."""


class BadRequestError(Error):
  """The parameters specified were in some way invalid."""


class NotImageError(Error):
  """The image data given is not recognizable as an image."""


class BadImageError(Error):
  """The image data is corrupt."""


class LargeImageError(Error):
  """The image data is too large to process."""


class InvalidBlobKeyError(Error):
  """The provided blob key was invalid."""

  def __init__(self, blob_key=None):
    """Constructor.

    Args:
      blob_key: The `blob_key` that is believed to be invalid. Set this argument
        to `None` if the blob key is unknown.
    """
    self._blob_key = blob_key

  def __str__(self):
    """Returns a string representation of this Error."""
    if self._blob_key:
      return "InvalidBlobKeyError: %s" % repr(self._blob_key)
    else:
      return "InvalidBlobKeyError"


class BlobKeyRequiredError(Error):
  """A blob key is required for this operation."""


class UnsupportedSizeError(Error):
  """The size that was specified is not supported by requested operation."""


class AccessDeniedError(Error):
  """The application does not have permission to access the image."""


class ObjectNotFoundError(Error):
  """The blob key referred to an object that does not exist."""


def _ToImagesError(error, blob_key=None):
  """Translates an application error to an Images error, if possible.

  Args:
    error: An `ApplicationError` to translate.
    blob_key: The `blob_key` that is used in the function that caused the error.
      Set this argument to `None` if the blob key is unknown.

  Returns:
    An Images API error if one is found; all other errors are returned as-is.
  """
  error_map = {
      images_service_pb2.ImagesServiceError.NOT_IMAGE:
          NotImageError,
      images_service_pb2.ImagesServiceError.BAD_IMAGE_DATA:
          BadImageError,
      images_service_pb2.ImagesServiceError.IMAGE_TOO_LARGE:
          LargeImageError,
      images_service_pb2.ImagesServiceError.INVALID_BLOB_KEY:
          InvalidBlobKeyError,
      images_service_pb2.ImagesServiceError.ACCESS_DENIED:
          AccessDeniedError,
      images_service_pb2.ImagesServiceError.OBJECT_NOT_FOUND:
          ObjectNotFoundError,
      images_service_pb2.ImagesServiceError.UNSPECIFIED_ERROR:
          TransformationError,
      images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA:
          BadRequestError,
  }

  error_code = error.application_error

  if error_code == images_service_pb2.ImagesServiceError.INVALID_BLOB_KEY:
    return InvalidBlobKeyError(blob_key)

  desired_exc = error_map.get(error_code, Error)
  return desired_exc(error.error_detail)


class Image(object):
  """Image object to manipulate."""

  def __init__(self, image_data=None, blob_key=None, filename=None):
    """Constructor.

    You can only specify one of the following arguments.

    Args:
      image_data: Byte of the image data.
      blob_key: A `BlobKey`, `BlobInfo`, string, or Unicode representation of
        the blob key of the blob that contains the image data.
      filename: String of the file name of a Google Storage file that contains
        the image data. Must be in the format `/gs/bucket_name/object_name`.

    Raises:
      NotImageError: If the given data is empty.
    """

    if not image_data and not blob_key and not filename:
      raise NotImageError("Empty image data.")
    if image_data and (blob_key or filename):
      raise NotImageError("Can only take one of image, blob key or filename.")
    if blob_key and filename:
      raise NotImageError("Can only take one of image, blob key or filename.")

    self._image_data = image_data
    if filename:
      self._blob_key = blobstore.create_gs_key(filename)
    else:
      self._blob_key = _extract_blob_key(blob_key)
    self._transforms = []
    self._width = None
    self._height = None
    self._format = None
    self._correct_orientation = UNCHANGED_ORIENTATION
    self._original_metadata = None

  @staticmethod
  def _byte_to_int(byte_):
    if six.PY2:
      return ord(byte_)
    else:
      return byte_

  def _check_transform_limits(self):
    """Ensures some simple limits on the number of transforms allowed.

    Raises:
      BadRequestError: If `MAX_TRANSFORMS_PER_REQUEST` transforms have already
          been requested for this image.
    """
    if len(self._transforms) >= MAX_TRANSFORMS_PER_REQUEST:
      raise BadRequestError("%d transforms have already been requested on this "
                            "image." % MAX_TRANSFORMS_PER_REQUEST)

  def _update_dimensions(self):
    """Updates the `width` and `height` fields of the image.

    Raises:
      NotImageError: If the image data is not an image.
      BadImageError: If the image data is corrupt.
    """
    if not self._image_data:
      raise NotImageError("Dimensions unavailable for blob key input")
    size = len(self._image_data)
    if size >= 6 and self._image_data.startswith(b"GIF"):
      self._update_gif_dimensions()
      self._format = GIF
    elif size >= 8 and self._image_data.startswith(b"\x89PNG\x0D\x0A\x1A\x0A"):
      self._update_png_dimensions()
      self._format = PNG
    elif size >= 2 and self._image_data.startswith(b"\xff\xD8"):
      self._update_jpeg_dimensions()
      self._format = JPEG
    elif (size >= 8 and (self._image_data.startswith(b"II\x2a\x00") or
                         self._image_data.startswith(b"MM\x00\x2a"))):
      self._update_tiff_dimensions()
      self._format = TIFF
    elif size >= 2 and self._image_data.startswith(b"BM"):
      self._update_bmp_dimensions()
      self._format = BMP
    elif size >= 4 and self._image_data.startswith(b"\x00\x00\x01\x00"):
      self._update_ico_dimensions()
      self._format = ICO
    elif (size >= 16 and (self._image_data.startswith(b"RIFF", 0, 4) and
                          self._image_data.startswith(b"WEBP", 8, 12) and
                          self._image_data.startswith(b"VP8 ", 12, 16))):
      self._update_webp_dimensions()
      self._format = WEBP

    elif (size >= 16 and (self._image_data.startswith(b"RIFF", 0, 4) and
                          self._image_data.startswith(b"WEBP", 8, 12) and
                          self._image_data.startswith(b"VP8X", 12, 16))):
      self._update_webp_vp8x_dimensions()
      self._format = WEBP
    else:
      raise NotImageError("Unrecognized image format")

  def _update_gif_dimensions(self):
    """Updates the `width` and `height` fields of the GIF image.

    Raises:
      BadImageError: If the image string is not a valid GIF image.
    """

    size = len(self._image_data)
    if size >= 10:
      self._width, self._height = struct.unpack("<HH", self._image_data[6:10])
    else:
      raise BadImageError("Corrupt GIF format")

  def _update_png_dimensions(self):
    """Updates the `width` and `height` fields of the PNG image.

    Raises:
      BadImageError: If the image string is not a valid PNG image.
    """


    size = len(self._image_data)
    if size >= 24 and self._image_data[12:16] == b"IHDR":
      self._width, self._height = struct.unpack(">II", self._image_data[16:24])
    else:
      raise BadImageError("Corrupt PNG format")

  def _update_jpeg_dimensions(self):
    """Updates the `width` and `height` fields of the JPEG image.

    Raises:
      BadImageError: If the image string is not a valid JPEG image.
    """

    size = len(self._image_data)
    offset = 2
    while offset < size:
      while offset < size and self._byte_to_int(
          self._image_data[offset]) != 0xFF:
        offset += 1
      while offset < size and self._byte_to_int(
          self._image_data[offset]) == 0xFF:
        offset += 1
      if (offset < size and
          self._byte_to_int(self._image_data[offset]) & 0xF0 == 0xC0 and
          self._byte_to_int(self._image_data[offset]) != 0xC4):
        offset += 4
        if offset + 4 <= size:
          self._height, self._width = struct.unpack(
              ">HH", self._image_data[offset:offset + 4])
          break
        else:
          raise BadImageError("Corrupt JPEG format")
      elif offset + 3 <= size:
        offset += 1
        offset += struct.unpack(">H", self._image_data[offset:offset + 2])[0]
      else:
        raise BadImageError("Corrupt JPEG format")
    if self._height is None or self._width is None:
      raise BadImageError("Corrupt JPEG format")

  def _update_tiff_dimensions(self):
    """Updates the `width` and `height` fields of the TIFF image.

    Raises:
      BadImageError: If the image string is not a valid TIFF image.
    """


    size = len(self._image_data)
    if self._image_data.startswith(b"II"):
      endianness = "<"
    else:
      endianness = ">"
    ifd_offset = struct.unpack(endianness + "I", self._image_data[4:8])[0]
    if ifd_offset + 14 <= size:
      ifd_size = struct.unpack(endianness + "H",
                               self._image_data[ifd_offset:ifd_offset + 2])[0]
      ifd_offset += 2
      for unused_i in range(0, ifd_size):
        if ifd_offset + 12 <= size:
          tag = struct.unpack(endianness + "H",
                              self._image_data[ifd_offset:ifd_offset + 2])[0]
          if tag == 0x100 or tag == 0x101:
            value_type = struct.unpack(
                endianness + "H",
                self._image_data[ifd_offset + 2:ifd_offset + 4])[0]
            if value_type == 3:
              format = endianness + "H"
              end_offset = ifd_offset + 10
            elif value_type == 4:
              format = endianness + "I"
              end_offset = ifd_offset + 12
            else:
              format = endianness + "B"
              end_offset = ifd_offset + 9
            if tag == 0x100:
              self._width = struct.unpack(
                  format, self._image_data[ifd_offset + 8:end_offset])[0]
              if self._height is not None:
                break
            else:
              self._height = struct.unpack(
                  format, self._image_data[ifd_offset + 8:end_offset])[0]
              if self._width is not None:
                break
          ifd_offset += 12
        else:
          raise BadImageError("Corrupt TIFF format")
    if self._width is None or self._height is None:
      raise BadImageError("Corrupt TIFF format")

  def _update_bmp_dimensions(self):
    """Updates the `width` and `height` fields of the BMP image.

    Raises:
      BadImageError: If the image string is not a valid BMP image.
    """




    size = len(self._image_data)
    if size >= 18:
      header_length = struct.unpack("<I", self._image_data[14:18])[0]
      if ((header_length == 40 or header_length == 108 or
           header_length == 124 or header_length == 64) and size >= 26):

        self._width, self._height = struct.unpack("<II",
                                                  self._image_data[18:26])
      elif header_length == 12 and size >= 22:
        self._width, self._height = struct.unpack("<HH",
                                                  self._image_data[18:22])
      else:
        raise BadImageError("Corrupt BMP format")
    else:
      raise BadImageError("Corrupt BMP format")

  def _update_ico_dimensions(self):
    """Updates the `width` and `height` fields of the ICO image.

    Raises:
      BadImageError: If the image string is not a valid ICO image.
    """
    size = len(self._image_data)
    if size >= 8:
      self._width, self._height = struct.unpack("<BB", self._image_data[6:8])

      if not self._width:
        self._width = 256
      if not self._height:
        self._height = 256
    else:
      raise BadImageError("Corrupt ICO format")

  def set_correct_orientation(self, correct_orientation):
    """Sets a flag to correct the image orientation based on image metadata.

    EXIF metadata within the image may contain a parameter that indicates its
    proper orientation. This value can equal 1 through 8, inclusive. Setting the
    orientation to 1 means that the image is in its "normal" orientation, i.e.,
    it should be viewed as it is stored. Normally, this orientation value has no
    effect on the behavior of the transformations. However, if you call this
    function with the `correct_orientation` argument, any orientation specified
    in the EXIF metadata will be corrected during the first transformation.

    Note:
        If `correct_orientation` is specified, but the image is already in
        portrait orientation, i.e., "taller" than it is "wide," no corrections
        will be made, since the camera has likely already corrected it.

    Whether the correction was requested or not, the orientation value in the
    transformed image is always cleared to indicate that no additional
    corrections of the returned image's orientation is necessary.

    Args:
      correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE`.

    Raises:
      BadRequestError: If the `correct_orientation` value is invalid.
    """
    if correct_orientation not in ORIENTATION_CORRECTION_TYPE:
      raise BadRequestError("Orientation correction must be in %s" %
                            ORIENTATION_CORRECTION_TYPE)
    self._correct_orientation = correct_orientation

  def _update_webp_dimensions(self):
    """Updates the width and height fields of the webp image."""

    size = len(self._image_data)



    if size < 30:
      raise BadImageError("Corrupt WEBP format")

    bits = self._byte_to_int(self._image_data[20]) | (
        self._byte_to_int(self._image_data[21]) << 8) | (
            self._byte_to_int(self._image_data[22]) << 16)

    key_frame = ((bits & 1) == 0)

    if not key_frame:
      raise BadImageError("Corrupt WEBP format")

    profile = (bits >> 1) & 7
    show_frame = (bits >> 4) & 1

    if profile > 3:
      raise BadImageError("Corrupt WEBP format")

    if show_frame == 0:
      raise BadImageError("Corrupt WEBP format")

    self._width, self._height = struct.unpack("<HH", self._image_data[26:30])

    if self._height is None or self._width is None:
      raise BadImageError("Corrupt WEBP format")

  def _update_webp_vp8x_dimensions(self):
    """Updates the width and height fields of a WEBP image with a vp8x chunk."""
    size = len(self._image_data)

    if size < 30:
      raise BadImageError("Corrupt WEBP format")

    self._width, self._height = struct.unpack("<II", self._image_data[24:32])

    if self._height is None or self._width is None:
      raise BadImageError("Corrupt WEBP format")

  def resize(self,
             width=0,
             height=0,
             crop_to_fit=False,
             crop_offset_x=0.5,
             crop_offset_y=0.5,
             allow_stretch=False):
    """Resizes the image, maintaining the aspect ratio.

    If both `width` and `height` are specified, the more restricting of the two
    values will be used when the image is resized. The maximum dimension allowed
    for both `width` and `height` is 4000 pixels.

    If both `width` and `height` are specified, and `crop_to_fit` is `True`, the
    less restricting of the two values will be used when the image is resized,
    and the image will be cropped to fit the specified size. In this case, the
    center of cropping can be adjusted by `crop_offset_x` and `crop_offset_y`.

    Args:
      width: Integer of the width in pixels to change the image width to.
      height: Integer of the height in pixels to change the image height to.
      crop_to_fit: If `True`, and both `width` and `height` are specified, the
        image is cropped after it is resized to fit the specified dimensions.
      crop_offset_x: Float value between 0.0 and 1.0. 0 is left, and 1 is right.
        The default value is 0.5, or the center of the image.
      crop_offset_y: Float value between 0.0 and 1.0. 0 is top, and 1 is bottom.
        The default value is 0.5, or the center of the image.
      allow_stretch: If `True`, and both `width` and `height` are specified, the
        image is stretched to fit the resize dimensions without maintaining the
        aspect ratio.

    Raises:
      TypeError: When `width` or `height` is not an integer or long type.
      BadRequestError: When the height or width is invalid, or if
          `MAX_TRANSFORMS_PER_REQUEST` transforms have already been
          requested on this image.
    """
    if (not isinstance(width, six.integer_types) or
        not isinstance(height, six.integer_types)):
      raise TypeError("Width and height must be integers.")
    if width < 0 or height < 0:
      raise BadRequestError("Width and height must be >= 0.")

    if not width and not height:
      raise BadRequestError("At least one of width or height must be > 0.")

    if width > 4000 or height > 4000:
      raise BadRequestError("Both width and height must be <= 4000.")

    if not isinstance(crop_to_fit, bool):
      raise TypeError("crop_to_fit must be boolean.")

    if crop_to_fit and not (width and height):
      raise BadRequestError("Both width and height must be > 0 when "
                            "crop_to_fit is specified.")

    if not isinstance(allow_stretch, bool):
      raise TypeError("allow_stretch must be boolean.")

    if allow_stretch and not (width and height):
      raise BadRequestError("Both width and height must be > 0 when "
                            "allow_stretch is specified.")

    self._validate_crop_arg(crop_offset_x, "crop_offset_x")
    self._validate_crop_arg(crop_offset_y, "crop_offset_y")

    self._check_transform_limits()

    transform = images_service_pb2.Transform()
    transform.width = width
    transform.height = height
    transform.crop_to_fit = crop_to_fit
    transform.crop_offset_x = crop_offset_x
    transform.crop_offset_y = crop_offset_y
    transform.allow_stretch = allow_stretch

    self._transforms.append(transform)

  def rotate(self, degrees):
    """Rotates an image a given number of degrees clockwise.

    Args:
      degrees: Integer value of the number of degrees that an image must be
        rotated. This value must be a multiple of 90.

    Raises:
      TypeError: When `degrees` is not an integer or long type.
      BadRequestError: When the `degrees` value is invalid or if
          `MAX_TRANSFORMS_PER_REQUEST` transforms have already been requested.
    """
    if not isinstance(degrees, six.integer_types):
      raise TypeError("Degrees must be integers.")

    if degrees % 90 != 0:
      raise BadRequestError("degrees argument must be multiple of 90.")


    degrees = degrees % 360




    self._check_transform_limits()

    transform = images_service_pb2.Transform()
    transform.rotate = degrees

    self._transforms.append(transform)

  def horizontal_flip(self):
    """Flips the image horizontally.

    Raises:
      BadRequestError: If `MAX_TRANSFORMS_PER_REQUEST` transforms have already
          been requested on the image.
    """
    self._check_transform_limits()

    transform = images_service_pb2.Transform()
    transform.horizontal_flip = True

    self._transforms.append(transform)

  def vertical_flip(self):
    """Flips the image vertically.

    Raises:
      BadRequestError: If `MAX_TRANSFORMS_PER_REQUEST` transforms have already
          been requested on the image.
    """
    self._check_transform_limits()
    transform = images_service_pb2.Transform()
    transform.vertical_flip = True

    self._transforms.append(transform)

  def _validate_crop_arg(self, val, val_name):
    """Validates the given value of a `Crop()` method argument.

    Args:
      val: Float value of the argument.
      val_name: String of the name of the argument.

    Raises:
      TypeError: If the arguments are not of type float.
      BadRequestError: When the bounding box is invalid.
    """
    if type(val) != float:
      raise TypeError("arg '%s' must be of type 'float'." % val_name)

    if not (0 <= val <= 1.0):
      raise BadRequestError("arg '%s' must be between 0.0 and 1.0 "
                            "(inclusive)" % val_name)

  def crop(self, left_x, top_y, right_x, bottom_y):
    """Crops the image.

    The four arguments are the scaling numbers that describe the bounding box
    that will crop the image.  The upper left point of the bounding box will
    be at (`left_x*image_width`, `top_y*image_height`), and the lower right
    point will be at (`right_x*image_width`, `bottom_y*image_height`).

    Args:
      left_x: Float value between 0.0 and 1.0, inclusive.
      top_y: Float value between 0.0 and 1.0, inclusive.
      right_x: Float value between 0.0 and 1.0, inclusive.
      bottom_y: Float value between 0.0 and 1.0, inclusive.

    Raises:
      TypeError: If the arguments are not of type float.
      BadRequestError: When the bounding box values are invalid or if
          `MAX_TRANSFORMS_PER_REQUEST` transforms have already been requested
          for this image.
    """
    self._validate_crop_arg(left_x, "left_x")
    self._validate_crop_arg(top_y, "top_y")
    self._validate_crop_arg(right_x, "right_x")
    self._validate_crop_arg(bottom_y, "bottom_y")

    if left_x >= right_x:
      raise BadRequestError("left_x must be less than right_x")
    if top_y >= bottom_y:
      raise BadRequestError("top_y must be less than bottom_y")

    self._check_transform_limits()

    transform = images_service_pb2.Transform()
    transform.crop_left_x = left_x
    transform.crop_top_y = top_y
    transform.crop_right_x = right_x
    transform.crop_bottom_y = bottom_y

    self._transforms.append(transform)

  def im_feeling_lucky(self):
    """Automatically adjusts the image's contrast and color levels.

    This is similar to the "I'm Feeling Lucky" button in Picasa.

    Raises:
      BadRequestError: If `MAX_TRANSFORMS_PER_REQUEST` transforms have already
          been requested for this image.
    """
    self._check_transform_limits()
    transform = images_service_pb2.Transform()
    transform.autolevels = True

    self._transforms.append(transform)

  def get_original_metadata(self):
    """Lists the metadata of the original image.

    Returns a dictionary of metadata extracted from the original image during
    `execute_transform`.

    Note:
        Some of the EXIF fields are processed; for example, fields with multiple
        values are returned as lists, rational types are returned as floats, GPS
        coordinates are already parsed to signed floats, etc.

    `ImageWidth` and `ImageLength` fields are corrected if they did not
    correspond to the actual dimensions of the original image.

    Returns:
      Dictionary of metadata with string keys. If `execute_transform` was
      called with `parse_metadata` set to `True`, this dictionary contains
      information about the various properties of the original image, such as
      dimensions, color profile, and properties from EXIF.

      Even if `parse_metadata` was `False` or the images did not have any
      metadata, the dictionary will contain a limited set of metadata, with
      least `ImageWidth` and `ImageLength` dimensions of the original image
      being returned.

      This call will return `None` if it is called before a successful
      `execute_transfrom` call.
    """
    return self._original_metadata

  def _set_imagedata(self, imagedata):
    """Populates an `ImageData` protocol buffer from this Image instance.

    Args:
        imagedata: An `ImageData` protocol buffer instance.
    """
    if self._blob_key:
      imagedata.content = b""
      imagedata.blob_key = self._blob_key
    else:
      imagedata.content = self._image_data

  def execute_transforms(self,
                         output_encoding=PNG,
                         quality=None,
                         parse_source_metadata=False,
                         transparent_substitution_rgb=None,
                         rpc=None):
    """Performs transformations on a given image.

    You can only execute one transformation per image per `execute_transforms()`
    call.

    Args:
      output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
      quality: A value between 1 and 100 to specify the quality of the encoding.
        This value is only used for JPEG and WEBP quality control.
      parse_source_metadata: When `True`, the metadata (EXIF) of the source
        image is parsed before any transformations. The results can be retrieved
        via `Image.get_original_metadata`.
      transparent_substitution_rgb: When transparent pixels are not supported in
        the destination image format, transparent pixels will be substituted for
        the specified color, which must be 32-bit RGB format.
      rpc: A UserRPC object.

    Returns:
      A string of the image data after the transformations have been
      performed on it.

    Raises:
      BadRequestError: When the request specifications are invalid.
      NotImageError: When the image data given is not an image.
      BadImageError: When the image data given is corrupt.
      LargeImageError: When the image data given is too large to process.
      InvalidBlobKeyError: When the blob key provided is invalid.
      TransformationError: When an error occurs during image manipulation.
      AccessDeniedError: When the blob key refers to a Google Storage object,
          and the application does not have permission to access the object.
      ObjectNotFoundError: When the blob key refers to an object that no
          longer exists.
      Error: All other error conditions.
    """
    rpc = self.execute_transforms_async(
        output_encoding=output_encoding,
        quality=quality,
        parse_source_metadata=parse_source_metadata,
        transparent_substitution_rgb=transparent_substitution_rgb,
        rpc=rpc)
    return rpc.get_result()

  def execute_transforms_async(self,
                               output_encoding=PNG,
                               quality=None,
                               parse_source_metadata=False,
                               transparent_substitution_rgb=None,
                               rpc=None):
    """Asynchronously performs transformations on a given image.

    Args:
      output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
      quality: A value between 1 and 100 to specify the quality of the encoding.
        This value is only used for JPEG & WEBP quality control.
      parse_source_metadata: When `True`, the metadata (EXIF) of the source
        image is parsed before any transformations. The results can be retrieved
        via `Image.get_original_metadata`.
      transparent_substitution_rgb: When transparent pixels are not supported in
        the destination image format, transparent pixels will be substituted for
        the specified color, which must be in 32-bit RGB format.
      rpc: A UserRPC object.

    Returns:
      A UserRPC object.

    Raises:
      BadRequestError: When the request specifications are invalid.
      NotImageError: When the image data given is not an image.
      BadImageError: When the image data given is corrupt.
      LargeImageError: When the image data given is too large to process.
      InvalidBlobKeyError: When the blob key provided is invalid.
      TransformationError: When an error occurs during image manipulation.
      AccessDeniedError: When the blob key refers to a Google Storage object,
          and the application does not have permission to access the object.
      ValueError: When `transparent_substitution_rgb` is not an integer.
      Error: All other error conditions.
    """
    if output_encoding not in OUTPUT_ENCODING_TYPES:
      raise BadRequestError("Output encoding type not in recognized set "
                            "%s" % OUTPUT_ENCODING_TYPES)

    if not self._transforms:
      raise BadRequestError("Must specify at least one transformation.")

    if transparent_substitution_rgb:
      if not isinstance(transparent_substitution_rgb, int):
        raise ValueError(
            "transparent_substitution_rgb must be a 32 bit integer")

    self.CheckValidIntParameter(quality, 1, 100, "quality")

    request = images_service_pb2.ImagesTransformRequest()
    response = images_service_pb2.ImagesTransformResponse()

    input_settings = request.input
    input_settings.correct_exif_orientation = self._correct_orientation

    if parse_source_metadata:
      input_settings.parse_metadata = True

    self._set_imagedata(request.image)

    for transform in self._transforms:
      request.transform.add().CopyFrom(transform)

    request.output.mime_type = output_encoding

    if ((output_encoding == JPEG or output_encoding == WEBP) and
        (quality is not None)):
      request.output.quality = quality

    if transparent_substitution_rgb:
      input_settings.transparent_substitution_rgb = transparent_substitution_rgb



    def execute_transforms_hook(rpc):
      """Checks success, handles exceptions, returns the converted RPC result.

      Args:
        rpc: A UserRPC object.

      Raises:
        See `execute_transforms_async` for more details.
      """
      try:
        rpc.check_success()
      except apiproxy_errors.ApplicationError as e:
        raise _ToImagesError(e, self._blob_key)
      self._image_data = rpc.response.image.content
      self._blob_key = None
      self._transforms = []
      if response.image.HasField("width"):
        self._width = rpc.response.image.width
      else:
        self._width = None
      if response.image.HasField("height"):
        self._height = rpc.response.image.height
      else:
        self._height = None
      self._format = None
      if response.source_metadata:
        self._original_metadata = json.loads(response.source_metadata)
      return self._image_data

    return _make_async_call(rpc, "Transform", request, response,
                            execute_transforms_hook, None)

  @property
  def width(self):
    """Retrieves the width of the image."""
    if self._width is None:
      self._update_dimensions()
    return self._width

  @property
  def height(self):
    """Retrieves the height of the image."""
    if self._height is None:
      self._update_dimensions()
    return self._height

  @property
  def format(self):
    """Retrieves the format of the image."""
    if self._format is None:
      self._update_dimensions()
    return self._format

  def histogram(self, rpc=None):
    """Calculates the histogram of the image.

    Args:
      rpc: A UserRPC object.

    Returns:
      3 256-element lists containing the number of occurences of each value of
      each color in the order RGB. You can learn more about color histograms
      at `Wikipedia`_.

    Raises:
      NotImageError: When the image data supplied is not an image.
      BadImageError: When the image data supplied is corrupt.
      LargeImageError: When the image data supplied is too large to process.
      Error: All other error conditions.

    .. _Wikipedia:
        http://en.wikipedia.org/wiki/Color_histogram

    """

    rpc = self.histogram_async(rpc)
    return rpc.get_result()

  def histogram_async(self, rpc=None):
    """Asynchronously calculates the histogram of the image.

    Args:
      rpc: An optional UserRPC object.

    Returns:
      rpc: A UserRPC object.

    Raises:
      NotImageError: When the image data supplied is not an image.
      BadImageError: When the image data supplied is corrupt.
      LargeImageError: When the image data supplied is too large to process.
      Error: All other error conditions.
    """
    request = images_service_pb2.ImagesHistogramRequest()
    response = images_service_pb2.ImagesHistogramResponse()

    self._set_imagedata(request.image)

    def get_histogram_hook(rpc):
      """Checks success, handles exceptions, returns the converted RPC result.

      Args:
        rpc: A UserRPC object.

      Returns:
        3 256-element lists containing the number of occurrences of each value
        of each color in the order RGB. To learn more about color histograms,
        see the `Wikipedia article`_ for color histograms.

      Raises:
        See `histogram_async` for more details.

      .. _Wikipedia article:
         https://en.wikipedia.org/wiki/Color_histogram
      """
      try:
        rpc.check_success()
      except apiproxy_errors.ApplicationError as e:
        raise _ToImagesError(e, self._blob_key)

      histogram_ = rpc.response.histogram
      return [histogram_.red, histogram_.green, histogram_.blue]

    return _make_async_call(rpc, "Histogram", request, response,
                            get_histogram_hook, None)

  @staticmethod
  def CheckValidIntParameter(parameter, min_value, max_value, name):
    """Checks that a parameter is an integer within the specified range."""

    if parameter is not None:
      if not isinstance(parameter, six.integer_types):
        raise TypeError("%s must be an integer." % name)
      if parameter > max_value or parameter < min_value:
        raise BadRequestError("%s must be between %s and %s." % name,
                              str(min_value), str(max_value))








def create_rpc(deadline=None, callback=None):
  """Creates an RPC object for use with the Images API.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
      is a system-specific deadline (typically 5 seconds).
    callback: Optional callable to invoke on completion.

  Returns:
    An `apiproxy_stub_map.UserRPC` object specialized for this service.
  """
  return apiproxy_stub_map.UserRPC("images", deadline, callback)


def _make_async_call(rpc, method, request, response, get_result_hook,
                     user_data):
  if rpc is None:
    rpc = create_rpc()
  rpc.make_call(method, request, response, get_result_hook, user_data)
  return rpc


def resize(image_data,
           width=0,
           height=0,
           output_encoding=PNG,
           quality=None,
           correct_orientation=UNCHANGED_ORIENTATION,
           crop_to_fit=False,
           crop_offset_x=0.5,
           crop_offset_y=0.5,
           allow_stretch=False,
           rpc=None,
           transparent_substitution_rgb=None):
  """Resizes a given image file while maintaining the aspect ratio.

  If both `width` and `height` are specified, the more restricting of the two
  values will be used when resizing the image. The maximum dimension allowed
  for both width and height is 4000 pixels.

  If both `width` and `height` are specified and `crop_to_fit` is `True`, the
  less restricting of the two values will be used when resizing and the image
  will be cropped to fit the specified size. In this case, the center of
  cropping can be adjusted by `crop_offset_x` and `crop_offset_y`.

  Args:
    image_data: String of the source image data.
    width: Integer of the width (in pixels) to change the image width to.
    height: Integer of the height (in pixels) to change the image height to.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    crop_to_fit: If `True`, and both `width` and `height` are specified, the
      image is cropped after being resized to fit the specified dimensions.
    crop_offset_x: Float value between 0.0 and 1.0. 0 is left and 1 is right.
      The default value is 0.5, or the center of image.
    crop_offset_y: Float value between 0.0 and 1.0. 0 is top and 1 is bottom.
      The default value is 0.5, or the center of image.
    allow_stretch: If `True`, and both `width` and `height` are specified, the
      image is stretched to fit the resized dimensions without maintaining the
      aspect ratio.
    rpc: Optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the resized image.

  Raises:
    TypeError: When `width` or `height` are not of integer or long type.
    BadRequestError: When the specified `width` or `height` is invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = resize_async(
      image_data,
      width=width,
      height=height,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      crop_to_fit=crop_to_fit,
      crop_offset_x=crop_offset_x,
      crop_offset_y=crop_offset_y,
      allow_stretch=allow_stretch,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def resize_async(image_data,
                 width=0,
                 height=0,
                 output_encoding=PNG,
                 quality=None,
                 correct_orientation=UNCHANGED_ORIENTATION,
                 crop_to_fit=False,
                 crop_offset_x=0.5,
                 crop_offset_y=0.5,
                 allow_stretch=False,
                 rpc=None,
                 transparent_substitution_rgb=None):
  """Asynchronously resizes an image file, maintaining the aspect ratio.

  If both `width` and `height` are specified, the more restricting of the two
  values will be used when resizing the image. The maximum dimension allowed
  for both width and height is 4000 pixels.

  If both `width` and `height` are specified and `crop_to_fit` is `True`, the
  less restricting of the two values will be used when resizing and the image
  will be cropped to fit the specified size. In this case, the center of
  cropping can be adjusted by `crop_offset_x` and `crop_offset_y`.

  Args:
    image_data: String of the source image data.
    width: Integer of the width in pixels to change the image width to.
    height: Integer of the height in pixels to change the image height to.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    crop_to_fit: If `True`, and both `width` and `height` are specified, the
      image is cropped after being resized to fit the specified dimensions.
    crop_offset_x: Float value between 0.0 and 1.0. 0 is left and 1 is right.
      The default value is 0.5, or the center of image.
    crop_offset_y: Float value between 0.0 and 1.0. 0 is top and 1 is bottom.
      The default value is 0.5, or the center of image.
    allow_stretch: If `True`, and both `width` and `height` are specified, the
      image is stretched to fit the resized dimensions without maintaining the
      aspect ratio.
    rpc: Optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object; call `get_result()` to obtain the result of the RPC.

  Raises:
    TypeError: When `width` or `height` are not of integer or long type.
    BadRequestError: When the specified `width` or `height` is invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.resize(
      width,
      height,
      crop_to_fit=crop_to_fit,
      crop_offset_x=crop_offset_x,
      crop_offset_y=crop_offset_y,
      allow_stretch=allow_stretch)
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def rotate(image_data,
           degrees,
           output_encoding=PNG,
           quality=None,
           correct_orientation=UNCHANGED_ORIENTATION,
           rpc=None,
           transparent_substitution_rgb=None):
  """Rotates an image a given number of degrees clockwise.

  Args:
    image_data: String of the source image data.
    degrees: A value from `ROTATE_DEGREE_VALUES`.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the rotated image.

  Raises:
    TypeError: When `degrees` is not of integer or long type.
    BadRequestError: When the specified `degrees` value is invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = rotate_async(
      image_data,
      degrees,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def rotate_async(image_data,
                 degrees,
                 output_encoding=PNG,
                 quality=None,
                 correct_orientation=UNCHANGED_ORIENTATION,
                 rpc=None,
                 transparent_substitution_rgb=None):
  """Asynchronously rotates an image a specified number of degrees clockwise.

  Args:
    image_data: String of the source image data.
    degrees: A value from `ROTATE_DEGREE_VALUES`.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object; call `get_result()` to complete the RPC and obtain the
    crop result.

  Raises:
    TypeError: When `degrees` is not of integer or long type.
    BadRequestError: When the specified `degrees` value is invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.rotate(degrees)
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def horizontal_flip(image_data,
                    output_encoding=PNG,
                    quality=None,
                    correct_orientation=UNCHANGED_ORIENTATION,
                    rpc=None,
                    transparent_substitution_rgb=None):
  """Flips the image horizontally.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the transformed image.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = horizontal_flip_async(
      image_data,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def horizontal_flip_async(image_data,
                          output_encoding=PNG,
                          quality=None,
                          correct_orientation=UNCHANGED_ORIENTATION,
                          rpc=None,
                          transparent_substitution_rgb=None):
  """Asynchronously flips the image horizontally.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object; call `get_result()` to complete the RPC and obtain the
    crop result.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.horizontal_flip()
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def vertical_flip(image_data,
                  output_encoding=PNG,
                  quality=None,
                  correct_orientation=UNCHANGED_ORIENTATION,
                  rpc=None,
                  transparent_substitution_rgb=None):
  """Flips the image vertically.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the transformed image.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = vertical_flip_async(
      image_data,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def vertical_flip_async(image_data,
                        output_encoding=PNG,
                        quality=None,
                        correct_orientation=UNCHANGED_ORIENTATION,
                        rpc=None,
                        transparent_substitution_rgb=None):
  """Asynchronously flips the image vertically.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object; call `get_result()` to complete the RPC and obtain the
    crop result.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.vertical_flip()
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def crop(image_data,
         left_x,
         top_y,
         right_x,
         bottom_y,
         output_encoding=PNG,
         quality=None,
         correct_orientation=UNCHANGED_ORIENTATION,
         rpc=None,
         transparent_substitution_rgb=None):
  """Crops the image.

  The four arguments are the scaling numbers that describe the bounding box
  that will crop the image.  The upper left point of the bounding box will be at
  (`left_x*image_width`, `top_y*image_height`), and the lower right point will
  be at (`right_x*image_width`, `bottom_y*image_height`).

  Args:
    image_data: String of the source image data.
    left_x: Float value between 0.0 and 1.0, inclusive.
    top_y: Float value between 0.0 and 1.0, inclusive.
    right_x: Float value between 0.0 and 1.0, inclusive.
    bottom_y: Float value between 0.0 and 1.0, inclusive.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the transformed image.

  Raises:
    TypeError: If the arguments are not of type float.
    BadRequestError: If the bounding box values are invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = crop_async(
      image_data,
      left_x,
      top_y,
      right_x,
      bottom_y,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def crop_async(image_data,
               left_x,
               top_y,
               right_x,
               bottom_y,
               output_encoding=PNG,
               quality=None,
               correct_orientation=UNCHANGED_ORIENTATION,
               rpc=None,
               transparent_substitution_rgb=None):
  """Asynchronously crops the given image.

  The four arguments are the scaling numbers that describe the bounding box
  that will crop the image.  The upper left point of the bounding box will be at
  (`left_x*image_width`, `top_y*image_height`), and the lower right point will
  be at (`right_x*image_width`, `bottom_y*image_height`).

  Args:
    image_data: String of the source image data.
    left_x: Float value between 0.0 and 1.0, inclusive.
    top_y: Float value between 0.0 and 1.0, inclusive.
    right_x: Float value between 0.0 and 1.0, inclusive.
    bottom_y: Float value between 0.0 and 1.0, inclusive.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object; call `get_result()` to complete the RPC and obtain the
    crop result.

  Raises:
    TypeError: If the args are not of type float.
    BadRequestError: When the bounding box values are invalid.
    Error: All other errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.crop(left_x, top_y, right_x, bottom_y)
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def im_feeling_lucky(image_data,
                     output_encoding=PNG,
                     quality=None,
                     correct_orientation=UNCHANGED_ORIENTATION,
                     rpc=None,
                     transparent_substitution_rgb=None):
  """Automatically adjusts image levels.

  This is similar to the "I'm Feeling Lucky" button in Picasa.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    Raw image data of the transformed image.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  rpc = im_feeling_lucky_async(
      image_data,
      output_encoding=output_encoding,
      quality=quality,
      correct_orientation=correct_orientation,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)
  return rpc.get_result()


def im_feeling_lucky_async(image_data,
                           output_encoding=PNG,
                           quality=None,
                           correct_orientation=UNCHANGED_ORIENTATION,
                           rpc=None,
                           transparent_substitution_rgb=None):
  """Asynchronously automatically adjusts image levels.

  This is similar to the "I'm Feeling Lucky" button in Picasa.

  Args:
    image_data: String of the source image data.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    correct_orientation: A value from `ORIENTATION_CORRECTION_TYPE` to indicate
      if orientation correction should be performed during the transformation.
    rpc: An optional UserRPC object.
    transparent_substitution_rgb: When transparent pixels are not supported in
      the destination image format, then transparent pixels will be substituted
      for the specified color, which must be in 32-bit RGB format.

  Returns:
    A UserRPC object.

  Raises:
    Error: All errors. See `Image.ExecuteTransforms` for more details.
  """
  image = Image(image_data)
  image.im_feeling_lucky()
  image.set_correct_orientation(correct_orientation)
  return image.execute_transforms_async(
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc,
      transparent_substitution_rgb=transparent_substitution_rgb)


def composite(inputs,
              width,
              height,
              color=0,
              output_encoding=PNG,
              quality=None,
              rpc=None):
  """Composites one or more images onto a canvas.

  Args:
    inputs: a list of tuples (`image_data`, `x_offset`, `y_offset`, `opacity`,
        `anchor`) where: - image_data String of the source image data. -
          x_offset X offset, in pixels from the anchor position. - y_offset Y
          offset, in piyels from the anchor position. - opacity Opacity of the
          image, specified as a float in range [0.0, 1.0] - anchor Anchoring
          point from `ANCHOR_POINTS`. The anchor point of the image is aligned
          with the same anchor point of the canvas. For example, `TOP_RIGHT`
          would place the top right corner of the image at the top right corner
          of the canvas, then apply the x and y offsets.
    width: Canvas width, in pixels.
    height: Canvas height, in pixels.
    color: Canvas background color, encoded as a 32-bit unsigned integer, where
      each color channel is represented by one byte in ARGB order.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used to control JPEG quality.
    rpc: Optional UserRPC object.

  Returns:
    A string that contains the image data of the composited image.

  Raises:
    TypeError: If `width`, `height`, `color`, `x_offset` or `y_offset` are not
        of type integer or long, or if `opacity` is not a float.
    BadRequestError: If more than `MAX_TRANSFORMS_PER_REQUEST` compositions
        have been requested, if the canvas width or height is greater than
        4000 or less than or equal to 0, if the color is invalid, if the opacity
        is outside the range [0,1] for any composition option, or if the anchor
        is invalid.
  """
  rpc = composite_async(
      inputs,
      width,
      height,
      color=color,
      output_encoding=output_encoding,
      quality=quality,
      rpc=rpc)
  return rpc.get_result()


def composite_async(inputs,
                    width,
                    height,
                    color=0,
                    output_encoding=PNG,
                    quality=None,
                    rpc=None):
  """Asynchronously composites one or more images onto a canvas.

  Args:
    inputs: A list of tuples (`image_data`, `x_offset`, `y_offset`, `opacity`,
        `anchor`), where: - image_data String of the source image data. -
          x_offset X offset, in pixels from the anchor position. - y_offset Y
          offset, in piyels from the anchor position. - opacity Opacity of the
          image, specified as a float in range [0.0, 1.0] - anchor Anchoring
          point from `ANCHOR_POINTS`. The anchor point of the image is aligned
          with the same anchor point of the canvas. For example, `TOP_RIGHT`
          would place the top right corner of the image at the top right corner
          of the canvas, then apply the x and y offsets.
    width: Canvas width, in pixels.
    height: Canvas height, in pixels.
    color: Canvas background color, encoded as a 32-bit unsigned integer, where
      each color channel is represented by one byte in ARGB order.
    output_encoding: A value from `OUTPUT_ENCODING_TYPES`.
    quality: A value between 1 and 100 to specify the quality of the encoding.
      This value is only used for JPEG quality control.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC object.

  Raises:
    TypeError: If `width`, `height`, `color`, `x_offset`, or `y_offset` are not
        an integer or long, or if `opacity` is not a float.
    BadRequestError: If more than `MAX_TRANSFORMS_PER_REQUEST` compositions
        have been requested, if the canvas width or height is greater than
        4000 or less than or equal to 0, if the color is invalid, if the opacity
        is outside the range [0,1] for any composition option, or the anchor is
        invalid.
  """
  if (not isinstance(width, six.integer_types) or
      not isinstance(height, six.integer_types) or
      not isinstance(color, six.integer_types)):
    raise TypeError("Width, height and color must be integers.")
  if output_encoding not in OUTPUT_ENCODING_TYPES:
    raise BadRequestError("Output encoding type '%s' not in recognized set "
                          "%s" % (output_encoding, OUTPUT_ENCODING_TYPES))

  if quality is not None:
    if not isinstance(quality, six.integer_types):
      raise TypeError("Quality must be an integer.")
    if quality > 100 or quality < 1:
      raise BadRequestError("Quality must be between 1 and 100.")

  if not inputs:
    raise BadRequestError("Must provide at least one input")
  if len(inputs) > MAX_COMPOSITES_PER_REQUEST:
    raise BadRequestError("A maximum of %d composition operations can be"
                          "performed in a single request" %
                          MAX_COMPOSITES_PER_REQUEST)

  if width <= 0 or height <= 0:
    raise BadRequestError("Width and height must be > 0.")
  if width > 4000 or height > 4000:
    raise BadRequestError("Width and height must be <= 4000.")

  if color > 0xffffffff or color < 0:
    raise BadRequestError("Invalid color")

  if color >= 0x80000000:
    color -= 0x100000000

  image_map = {}

  request = images_service_pb2.ImagesCompositeRequest()
  response = images_service_pb2.ImagesTransformResponse()
  for (image, x, y, opacity, anchor) in inputs:
    if not image:
      raise BadRequestError("Each input must include an image")
    if (not isinstance(x, six.integer_types) or
        not isinstance(y, six.integer_types) or
        not isinstance(opacity, (float))):
      raise TypeError("x_offset, y_offset must be integers and opacity must"
                      "be a float")
    if x > 4000 or x < -4000:
      raise BadRequestError("xOffsets must be in range [-4000, 4000]")
    if y > 4000 or y < -4000:
      raise BadRequestError("yOffsets must be in range [-4000, 4000]")
    if opacity < 0 or opacity > 1:
      raise BadRequestError("Opacity must be in the range 0.0 to 1.0")
    if anchor not in ANCHOR_TYPES:
      raise BadRequestError("Anchor type '%s' not in recognized set %s" %
                            (anchor, ANCHOR_TYPES))
    if image not in image_map:
      image_map[image] = len(request.image)

      if isinstance(image, Image):
        image._set_imagedata(request.image.add())
      else:
        request.image.add().content = image

    option = request.options.add()
    option.x_offset = x
    option.y_offset = y
    option.opacity = opacity
    option.anchor = anchor
    option.source_index = image_map[image]

  request.canvas.output.mime_type = output_encoding
  request.canvas.width = width
  request.canvas.height = height
  request.canvas.color = color

  if ((output_encoding == JPEG or output_encoding == WEBP) and
      (quality is not None)):
    request.canvas.output.quality = quality

  def composite_hook(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    Args:
      rpc: A UserRPC object.

    Returns:
      Images bytes of the composite image.

    Raises:
      See `composite_async` for more details.
    """
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as e:
      raise _ToImagesError(e)
    return rpc.response.image.content

  return _make_async_call(rpc, "Composite", request, response, composite_hook,
                          None)


def histogram(image_data, rpc=None):
  """Calculates the histogram of the given image.

  Args:
    image_data: String of the source image data.
    rpc: An optional UserRPC object.

  Returns:
    3 256-element lists containing the number of occurences of each value of
    each color in the order RGB.

  Raises:
    NotImageError: If the supplied image data is not an image.
    BadImageError: If the supplied image data is corrupt.
    LargeImageError: If the supplied image data is too large to process.
    Error: All other errors.
  """
  rpc = histogram_async(image_data, rpc=rpc)
  return rpc.get_result()


def histogram_async(image_data, rpc=None):
  """Calculates the histogram of the given image - async version.

  Args:
    image_data: String of the source image data.
    rpc: An optional UserRPC object.

  Returns:
    An UserRPC object.

  Raises:
    NotImageError: If the supplied image data is not an image.
    BadImageError: If the supplied image data is corrupt.
    LargeImageError: If the supplied image data is too large to process.
    Error: All other errors.
  """
  image = Image(image_data)
  return image.histogram_async(rpc)


IMG_SERVING_SIZES_LIMIT = 3200


IMG_SERVING_SIZES = [
    32, 48, 64, 72, 80, 90, 94, 104, 110, 120, 128, 144, 150, 160, 200, 220,
    288, 320, 400, 512, 576, 640, 720, 800, 912, 1024, 1152, 1280, 1440, 1600
]


IMG_SERVING_CROP_SIZES = [32, 48, 64, 72, 80, 104, 136, 144, 150, 160]


def get_serving_url(blob_key,
                    size=None,
                    crop=False,
                    secure_url=None,
                    filename=None,
                    rpc=None):
  """Obtains a URL that will serve the underlying image.

  This URL is served by a high-performance dynamic image serving infrastructure.
  This URL format also allows dynamic resizing and cropping with certain
  restrictions. To dynamically resize and crop, specify size and crop arguments,
  or simply append options to the end of the default URL obtained via this call.

  Example::

      get_serving_url -> "http://lh3.ggpht.com/SomeCharactersGoesHere"


  To get a 32-pixel-sized version (aspect-ratio preserved), append `=s32` to the
  URL::

      http://lh3.ggpht.com/SomeCharactersGoesHere=s32


  To get a 32-pixel cropped version, append `=s32-c`::

      http://lh3.ggpht.com/SomeCharactersGoesHere=s32-c


  Available sizes are any integer in the range [0, 1600] and is available as
  `IMG_SERVING_SIZES_LIMIT`.

  Args:
    blob_key: The `BlobKey`, `BlobInfo`, string, or Unicode representation of
      `BlobKey` of the blob whose URL you require.
    size: Integer of the size of resulting images.
    crop: Boolean value. `True` requests a cropped image, while `False` requests
      a resized image.
    secure_url: Boolean value. `True` requests a `https` URL, while `False`
      requests a `http` URL.
    filename: The file name of a Google Storage object whose URL you require.
    rpc: Optional UserRPC object.

  Returns:
    A URL string.

  Raises:
    BlobKeyRequiredError: If a blob key was not specified in the constructor.
    UnsupportedSizeError: If you specified an invalid size parameter.
    BadRequestError: If `crop` and `size` are present in the wrong
        combination, or if `blob_key` and `filename` are both specified.
    TypeError: If `secure_url` is not a boolean type.
    AccessDeniedError: If `blob_key` refers to a Google Storage object and
        the application does not have permission to access the object.
    ObjectNotFoundError: If `blob_key` refers to an object that doesn't
        exist.
  """
  rpc = get_serving_url_async(blob_key, size, crop, secure_url, filename, rpc)
  return rpc.get_result()


def get_serving_url_async(blob_key,
                          size=None,
                          crop=False,
                          secure_url=None,
                          filename=None,
                          rpc=None):
  """Asynchronously obtains a URL that will serve the underlying image.

  This URL is served by a high-performance dynamic image serving infrastructure.
  This URL format also allows dynamic resizing and cropping with certain
  restrictions. To dynamically resize and crop, specify size and crop arguments,
  or simply append options to the end of the default URL obtained via this call.

  Example::

      get_serving_url -> "http://lh3.ggpht.com/SomeCharactersGoesHere"


  To get a 32-pixel-sized version (aspect-ratio preserved), append `=s32` to the
  URL::

      http://lh3.ggpht.com/SomeCharactersGoesHere=s32


  To get a 32-pixel cropped version, append `=s32-c`::

      http://lh3.ggpht.com/SomeCharactersGoesHere=s32-c


  Available sizes are any integer in the range [0, 1600] and is available as
  `IMG_SERVING_SIZES_LIMIT`.

  Args:
    blob_key: The `BlobKey`, `BlobInfo`, string, or Unicode representation of
      the `BlobKey` of blob whose URL you require.
    size: Integer of the size of resulting images.
    crop: Boolean value. `True` requests a cropped image, while `False` requests
      a resized image.
    secure_url: Boolean value. `True` requests a `https` URL, while `False`
      requests a `http` URL.
    filename: The file name of a Google Storage object whose URL you require.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC whose result will be a string that is the serving URL.

  Raises:
    BlobKeyRequiredError: If a blob key was not specified in the constructor.
    UnsupportedSizeError: If you specified an invalid size parameter.
    BadRequestError: If `crop` and `size` are present in the wrong
        combination, or if `blob_key` and `filename` are both specified.
    TypeError: If `secure_url` is not a boolean type.
    AccessDeniedError: If `blob_key` refers to a Google Storage object and
        the application does not have permission to access the object.
  """
  if not blob_key and not filename:
    raise BlobKeyRequiredError(
        "A Blobkey or a filename is required for this operation.")

  if crop and not size:
    raise BadRequestError("Size should be set for crop operation")

  if size is not None and (size > IMG_SERVING_SIZES_LIMIT or size < 0):
    raise UnsupportedSizeError("Unsupported size")

  if secure_url and not isinstance(secure_url, bool):
    raise TypeError("secure_url must be boolean.")

  if filename and blob_key:
    raise BadRequestError("Cannot specify a blob_key and a filename.")

  if filename:
    _blob_key = blobstore.create_gs_key(filename)
    readable_blob_key = filename
  else:
    _blob_key = _extract_blob_key(blob_key)
    readable_blob_key = blob_key

  request = images_service_pb2.ImagesGetUrlBaseRequest()
  response = images_service_pb2.ImagesGetUrlBaseResponse()

  request.blob_key = _blob_key

  if secure_url:
    request.create_secure_url = secure_url

  def get_serving_url_hook(rpc):
    """Checks success, handles exceptions, and returns a converted RPC result.

    Args:
      rpc: A UserRPC object.

    Returns:
      The URL to serve the image.

    Raises:
      See `get_serving_url` for more details.
    """
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as e:
      raise _ToImagesError(e, readable_blob_key)

    url = rpc.response.url

    if size is not None:
      url += "=s%s" % size
    if crop:
      url += "-c"

    return url

  return _make_async_call(rpc, "GetUrlBase", request, response,
                          get_serving_url_hook, None)


def delete_serving_url(blob_key, rpc=None):
  """Deletes a serving URL created for `blob_key` using `get_serving_url`.

  Args:
    blob_key: The `BlobKey`, `BlobInfo`, string, or Unicode representation of
      the BlobKey of a blob with an existing URL that you want to delete.
    rpc: Optional UserRPC object.

  Raises:
    BlobKeyRequiredError: If no `blob_key` was specified.
    InvalidBlobKeyError: If the `blob_key` supplied was invalid.
    Error: There was a generic error deleting the serving URL.
  """
  rpc = delete_serving_url_async(blob_key, rpc)
  rpc.get_result()


def delete_serving_url_async(blob_key, rpc=None):
  """Deletes a serving URL created using get_serving_url - async version.

  Args:
    blob_key: `BlobKey`, `BlobInfo`, string, or unicode representation of the
      BlobKey of a blob with an existing URL that you want to delete.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC object.

  Raises:
    BlobKeyRequiredError: If no `blob_key` was specified.
    InvalidBlobKeyError: If the `blob_key` supplied was invalid.
    Error: There was a generic error deleting the serving URL.
  """
  if not blob_key:
    raise BlobKeyRequiredError("A Blobkey is required for this operation.")

  request = images_service_pb2.ImagesDeleteUrlBaseRequest()
  response = images_service_pb2.ImagesDeleteUrlBaseResponse()

  request.blob_key = _extract_blob_key(blob_key)

  def delete_serving_url_hook(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    Args:
      rpc: A UserRPC object.

    Raises:
      See `delete_serving_url_async` for more details.
    """
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as e:
      raise _ToImagesError(e, blob_key)

  return _make_async_call(rpc, "DeleteUrlBase", request, response,
                          delete_serving_url_hook, None)


def _extract_blob_key(blob):
  """Extracts a Unicode blob key from a string, `BlobKey`, or `BlobInfo`.

  Args:
    blob: The string, Unicode, `BlobKey`, or `BlobInfo` that contains the blob
      key.

  Returns:
    Unicode string representing the blob key.
  """
  if isinstance(blob, bytes):
    return blob.decode("utf-8")
  elif isinstance(blob, BlobKey):
    return str(blob)
  elif blob.__class__.__name__ == "BlobInfo":



    return str(blob.key())


  return blob
